import io
import logging
from typing import List, Dict, Union

import pandas as pd
from django.db import models
from packaging.version import Version
from pyarrow import BufferReader

from hostore.utils.timeseries import ts_combine_first

logger = logging.getLogger(__name__)
# below this version of pandas we must remove datetime index from the serie to be feathered
MIN_PANDAS_VERSION_FEATHER_SAVE_DATETIME_INDEX = '2.2.0'

class TimeseriesStore(models.Model):
    data = models.BinaryField(blank=True, null=True)

    class Meta:
        abstract = True
        # must be customized
        constraints = [models.UniqueConstraint(fields=['data'], name='%(app_label)s_%(class)s_data_orig_ctn'), ]
        indexes = [models.Index(fields=['data']), ]  # must be customized

    @property
    def decoded_ts_data(self) -> dict:
        """
        Returns a dict of all entries and timeserie ('data' key)
        """
        entry_dict = vars(self)

        # decode
        reader = BufferReader(entry_dict['data'])
        df = pd.read_feather(reader)
        # fix feather index
        if 'index' in df.columns:
            df.set_index('index', inplace=True)
        entry_dict['data'] = df.iloc[:, 0]
        return entry_dict

    @staticmethod
    def encode_serie(ds):
        """
        Encode and compress serie
        """
        # pandas to feather
        df = ds.to_frame(name='name')

        # fix feather index
        if Version(pd.__version__) < Version(MIN_PANDAS_VERSION_FEATHER_SAVE_DATETIME_INDEX):
            # remove index from serie
            df.reset_index(inplace=True, names=['index'])

        buf = io.BytesIO()
        df.to_feather(buf, compression='lz4')
        data = buf.getvalue()
        return data

    @classmethod
    def get_ts(cls, ts_attributes: dict, flat=False, _qs=None) -> Union[pd.Series, List[Dict]]:
        """
        Get the timeseries matching ts_attributes

        Args:
            ts_attributes: dict : specify attributes to get
            flat: bool : whether the query must return only one results (and return only data serie). when using flat
            option, there must be one and only one matching result
            _qs : private argument to pass ts_attributes queryset
        Returns:
            List[Dict]
        """
        full_key = repr(ts_attributes)
        logger.debug(f'GET key {full_key} in cache')

        if _qs is not None:
            qs = _qs
        else:
            qs = cls.objects.filter(**ts_attributes)

        entries = []
        for entry in qs:
            entries.append(entry.decoded_ts_data)

        logger.debug(f'GET key {full_key} in cache DONE')
        if flat:
            if len(entries) == 1:
                return entries[0]['data']
            elif len(entries) == 0:
                raise ValueError(f'No serie found for key {full_key}')
            else:
                raise ValueError(f'Multiple series found for key {full_key}')
        else:
            return entries

    @classmethod
    def set_ts(cls, ts_attributes: dict, ds_ts: pd.Series, update=False, replace=False) -> None:
        """
        Set one timeserie with ts_attributes keys

        Args:
            ts_attributes: dict : specify attributes to set
            ds_ts: pd.Series
            update: bool : if True, allow to update an existing ts (combine first existing and new one)
            replace: bool : if True, allow to replace an existing ts
        Returns:
            None
        """
        if update and replace:
            raise ValueError('update and replace are mutually exclusive')

        if ts_attributes is None:
            raise ValueError(f'ts_attributes is None')

        full_key = repr(ts_attributes)
        logger.debug(f'SET key {full_key} in cache')

        if not isinstance(ds_ts, pd.Series):
            raise ValueError(f'Cannot cache value of type {type(ds_ts)}, only pd.Series are accepted')

        if ds_ts.isnull().all():
            logger.warning(f'CACHE : Key {full_key} is ignored because data is null')
            return

        qs = cls.objects.filter(**ts_attributes)
        if qs.exists():
            if update:
                ds_ts_existing = cls.get_ts(ts_attributes, flat=True, _qs=qs)
                ds_ts_saved = ts_combine_first([ds_ts, ds_ts_existing])
            elif replace:
                cls.clear(ts_attributes)
                ds_ts_saved = ds_ts
            else:
                raise ValueError(f'Trying save over existing ts without update or replace option: {ts_attributes}')
        else:
            ds_ts_saved = ds_ts

        cls.objects.update_or_create(defaults=dict(data=cls.encode_serie(ds_ts_saved)), **ts_attributes)
        logger.debug(f'SET key {full_key} in cache DONE')

    @classmethod
    def clear(cls, custom_filters) -> None:
        """
        Method to clear multiple prm from cache
        Args:
            custom_filters: dict
        Returns:
            None
        """
        qs = cls.objects.filter(**custom_filters)
        qs.delete()



import base64, datetime as dt
from unittest.mock import patch

import requests
from django.db import connection, models, IntegrityError
from django.test import TransactionTestCase, override_settings
from django.urls import path, include
from rest_framework.routers import SimpleRouter
from rest_framework.test import APIClient, RequestsClient
import pandas as pd
import numpy as np
import lz4.frame

from hostore.models import TimeseriesChunkStore
from hostore.tests.test_chunk_timeserie_store import assert_series_equal
from hostore.utils.ts_sync import TimeseriesChunkStoreSyncViewSet, TimeseriesChunkStoreSyncClient


# ------------------------------------------------------------------
#  Modèles “remote” et “local” (mêmes champs, tables distinctes)
# ------------------------------------------------------------------
class ServerStore(TimeseriesChunkStore):
    """Table côté ‘serveur’"""
    version = models.IntegerField()
    kind    = models.CharField(max_length=20)
    ALLOW_CLIENT_SERVER_SYNC = True


class ClientStore(TimeseriesChunkStore):
    """Table côté ‘client’"""
    version = models.IntegerField()
    kind    = models.CharField(max_length=20)
    ALLOW_CLIENT_SERVER_SYNC = True


# ------------------------------------------------------------------
#  ViewSet serveur (lié au ServerStore)
# ------------------------------------------------------------------
RemoteSyncViewSet = TimeseriesChunkStoreSyncViewSet.as_factory(ServerStore)
router = SimpleRouter()
router.register("sync/year", RemoteSyncViewSet, basename="sync-year")
urlpatterns = [path("", include(router.urls))]      #   (1)  renommé → urlpatterns

test_urls = [path("", include(router.urls))]


# ------------------------------------------------------------------
#  Test d’intégration complet
# ------------------------------------------------------------------
@override_settings(ROOT_URLCONF=__name__)           #   (2)  on passe le module courant
class SyncIntegrationTestCase(TransactionTestCase):
    reset_sequences = True

    def setUp(self):
        # crée les tables manquantes
        with connection.schema_editor(atomic=False) as se:
            for mdl in (ServerStore, ClientStore):
                if mdl._meta.db_table not in connection.introspection.table_names():
                    se.create_model(mdl)
        ServerStore.objects.all().delete(_disable_sync_safety=True)
        self.sync_client = TimeseriesChunkStoreSyncClient(endpoint='http://testserver/sync/year', store_model=ClientStore)
        ClientStore.objects.all().delete(_disable_sync_safety=True)
        self.req_client = RequestsClient()
        self.req_client.base_url = "http://testserver/"
        # self.client = APIClient()


    def _sync(self, filters=None, **kwargs):
        with (
            patch.object(requests, "get",  self.req_client.get),
            # patch.object(requests, "post", self.req_client.post),
            # patch.object(requests, "put",  self.req_client.put),
        ):
            call_kwargs = {"batch": 20, "filters": filters, "page_size": 4}
            call_kwargs.update(kwargs)
            return self.sync_client.pull(**call_kwargs)

    # --------------------------------------------------------------
    def test_full_sync_and_update(self):
        # =========== TEST BASE
        # seed serveur
        sere_a1_v1 = self._make_series("2025-01-01", 24 * 31 * 4)
        sere_a2_v1 = self._make_series("2026-01-05", 24 * 50)
        series = (
            ({"version": 1, "kind": "A"}, sere_a1_v1),
            ({"version": 2, "kind": "A"}, sere_a2_v1),
            ({"version": 2, "kind": "B"}, self._make_series("2024-01-05", 24 * 50)),
            ({"version": 2, "kind": "c"}, self._make_series("2000-01-05", 24 * 50)),
        )
        for attr, serie in series:
            ServerStore.set_ts(attr, serie, replace=True)
        self._sync()
        self._assert_stores_equal()

        # =========== TEST UPDATE + SYNC PARTIEL / TOTAL
        # update server
        serie_a1_v2 = self._make_series("2025-02-28", 24 * 31 * 4)
        serie_a2_v2 = self._make_series("2026-03-01", 24 * 31 * 4)
        series2 = (
            ({"version": 1, "kind": "A"}, serie_a1_v2),
            ({"version": 2, "kind": "A"}, serie_a2_v2),
        )
        for attr, serie in series2:
            ServerStore.set_ts(attr, serie, update=True)

        # sync kind B : no changes for kind A on client side
        n_fetch, n_del = self._sync(filters={"kind": "B"})
        self.assertEqual(n_fetch, 1)  # gte => 1 chunk
        self.assertEqual(n_del, 0)
        self._assert_stores_equal(filters={"kind": "B"})

        # check no changes for kind A on client side
        for attr, serie in series:
            got = ClientStore.get_ts(attr)
            got = got[got.notnull()]
            assert_series_equal(got, serie)

        # sync all
        self._sync()
        self._assert_stores_equal()

        # manual check of content, in duplicate with previous _assert_stores_equal
        series_verif = [
            ({"version": 1, "kind": "A"}, serie_a1_v2.combine_first(sere_a1_v1)),
            ({"version": 2, "kind": "A"}, serie_a2_v2.combine_first(sere_a2_v1)),
            series[2], series[3],
        ]
        for attr, serie in series_verif:
            got = ClientStore.get_ts(attr)
            got = got[got.notnull()]
            assert_series_equal(got, serie)

        # =========== TEST SUPPRESSION
        # delete one object
        first = ServerStore.objects.filter(version=1).first()
        first.delete(keep_tracking=True)
        first.refresh_from_db()
        self.assertEqual(first.is_deleted, True)
        # delete qs
        ServerStore.objects.filter(version=1).delete(keep_tracking=True)

        self._sync()
        self._assert_stores_equal()
        self.assertFalse(ClientStore.objects.filter(version=1, is_deleted=False).exists())

        # =========== TEST REPLACE
        # replace server
        series3 = (
            ({"version": 1, "kind": "A"}, self._make_series("2025-03-01", 24 * 31 * 4)),
            ({"version": 2, "kind": "A"}, self._make_series("2026-04-01", 24 * 31 * 4)),
            ({"version": 2, "kind": "new_one"}, self._make_series("2025-04-01", 24 * 31)),
        )
        for attr, serie in series3:
            ServerStore.set_ts(attr, serie, replace=True)

        self._sync()
        self._assert_stores_equal()

        # =========== TEST DELETE SET MANY
        # delete without tracking on server side WILL lead to inconsistent results client side
        # => forbidden
        with self.assertRaises(ValueError):
            ServerStore.objects.all().delete(keep_tracking=False)

        # tracked model => you cannot set
        attr, serie = series[0]
        with self.assertRaises(ValueError):
            ServerStore.set_ts(attr, serie)

        # tracked model => you cannot set_many_ts
        mapping = {
            (k['version'], k['kind']): ds for k, ds in series3
        }
        with self.assertRaises(ValueError):
            ServerStore.set_many_ts(mapping, keys=("version", "kind"))

        # some legal update of series
        series4 = (
            ({"version": 1, "kind": "A"}, self._make_series("2025-03-01", 24 * 31 * 6)),
            ({"version": 2, "kind": "A"}, self._make_series("2026-04-01", 24 * 31 * 4)),
            ({"version": 2, "kind": "new_one"}, self._make_series("2025-04-01", 24 * 31 * 7)),
            ({"version": 2, "kind": "new_one2"}, self._make_series("2025-04-01", 24 * 31 * 7)),
            ({"version": 42, "kind": "d"}, self._make_series("1780-04-01", 24 * 31 * 7)),
        )
        for attr, serie in series4:
            ServerStore.set_ts(attr, serie, update=True)
        self._sync()
        self._assert_stores_equal()

        # =========== TEST DELETE FILTER
        ServerStore.objects.filter(kind="A").delete(keep_tracking=True)
        self._sync()
        self._assert_stores_equal()

        # =========== TEST DELETE ALL
        ServerStore.objects.all().delete(keep_tracking=True)
        self._sync()
        self._assert_stores_equal()


    # --------------------------------------------------------------
    @staticmethod
    def _make_series(start, periods, freq="1h", tz="Europe/Paris"):
        idx = pd.date_range(start=start, periods=periods, freq=freq, tz=tz)
        np.random.seed(0)
        return pd.Series(np.random.randn(periods), index=idx)

    def _assert_stores_equal(self, filters=None):
        filters = filters or {}
        seen_local = {
            (key_dict['version'], key_dict['kind']): serie
            for serie, key_dict in ClientStore.yield_many_ts(filters)
        }
        seen_remote = {
            (key_dict['version'], key_dict['kind']): serie
            for serie, key_dict in ServerStore.yield_many_ts(filters)
        }

        self.assertEqual(seen_local.keys(), seen_remote.keys())
        for k in seen_local.keys():
            assert_series_equal(seen_local[k], seen_remote[k])

    def test_partial_sync_crash_resumes(self):
        # Seed server with two series and sync once
        series = (
            ({"version": 1, "kind": "A"}, self._make_series("2025-01-01", 24)),
            ({"version": 2, "kind": "A"}, self._make_series("2025-01-02", 24)),
        )
        for attrs, serie in series:
            ServerStore.set_ts(attrs, serie, replace=True)
        self._sync()

        # Update both series on server to create two updates
        updated_series = (
            ({"version": 1, "kind": "A"}, self._make_series("2025-01-01", 48)),
            ({"version": 2, "kind": "A"}, self._make_series("2025-01-02", 48)),
        )
        for attrs, serie in updated_series:
            ServerStore.set_ts(attrs, serie, update=True)

        original_import = ClientStore.import_chunks.__func__
        call_count = {"value": 0}

        def crashing_import(cls, payload):
            call_count["value"] += 1
            original_import(cls, payload)
            if call_count["value"] == 1:
                raise RuntimeError("boom")

        with patch.object(ClientStore, "import_chunks", new=classmethod(crashing_import)):
            with self.assertRaises(RuntimeError):
                self._sync(batch=1)

        # Ensure only one update made it through
        self.assertEqual(call_count["value"], 1)

        # Second sync (without crashing) should resume and finish importing
        fetch, delete = self._sync()
        self.assertEqual(fetch, 2)
        self.assertEqual(delete, 0)
        self._assert_stores_equal()

    def test_pull_uses_filters_for_since(self):
        series = (
            ({"version": 1, "kind": "A"}, self._make_series("2025-01-01", 24)),
            ({"version": 1, "kind": "B"}, self._make_series("2025-01-01", 24)),
        )
        for attrs, serie in series:
            ServerStore.set_ts(attrs, serie, replace=True)

        called_with = {}

        original_last_updated = ClientStore.last_updated_at.__func__

        def wrapped_last_updated(cls, filters=None):
            called_with["filters"] = filters
            return original_last_updated(cls, filters)

        with patch.object(ClientStore, "last_updated_at", new=classmethod(wrapped_last_updated)):
            self._sync(filters={"kind": "A"})

        self.assertEqual(called_with["filters"], {"kind": "A"})


def _display_table_content(table, filters=None):
    """ debug tool to analyse table content (without data)"""
    filters = filters or {}
    df = pd.DataFrame(table.objects.filter(**filters).values())
    if 'data' in df.columns:
        df.drop(['data'], inplace=True, axis=1)
    print(f'Raw table content for {table.__name__}:')
    print(df.sort_values(list(df.columns)).to_markdown())


import io
import logging
import zipfile

import datetime as dt
from gettext import gettext

import pandas as pd
from django.contrib import admin, messages
from django.http import HttpResponse

logger = logging.getLogger(__name__)


class CompressedExport:
    def __init__(self):
        self.attachment_files = []

    def append_df_attach(self, df, filename):
        """
        Append Dataframe as CSV
        """
        buffer = io.BytesIO()
        df.to_csv(buffer, **dict(sep=';', decimal='.', float_format='%.8f'))
        buffer.seek(0)
        self.attachment_files.append((filename, buffer.getvalue()))

    def make_zip_binary(self):
        # Create a bytes buffer for the ZIP file
        zip_buffer = io.BytesIO()
        # Create a ZIP file in memory and add the CSV data
        with zipfile.ZipFile(zip_buffer, 'a', zipfile.ZIP_DEFLATED, False) as zip_file:
            for attach in self.attachment_files:
                zip_file.writestr(attach[0], attach[1])
        zip_buffer.seek(0)  # Go back to the start of the buffer
        return zip_buffer


@admin.action(description="Download selected timeseries")
def download_timeseries_from_store(modeladmin, request, queryset):
    holc_ts_qs = queryset.all()
    if not holc_ts_qs.exists():
        modeladmin.message_user(
            request,
            gettext("Please select at least one object"),
            messages.WARNING,
        )
        return
    datas = []
    ts_obj = None
    for ts_obj in holc_ts_qs:
        obj_content = ts_obj.decoded_ts_data
        ds = obj_content.pop('data')
        obj_content.pop('id')

        obj_content = {k: v for k, v in obj_content.items() if k[0] != '_'}
        datas.append((ds, obj_content))

    exp = CompressedExport()
    summary_data = []
    for ii, (ds, obj_content) in enumerate(datas):
        # Application du format
        filename = f'export_serie_{ii}.csv'
        df = ds.to_frame(name='data')
        exp.append_df_attach(df, filename)
        summary_data.append({'filename': filename, **obj_content})

    # Append summary
    df_summary = pd.DataFrame(summary_data)
    exp.append_df_attach(df_summary, f'content_summary.csv')

    # Make binary
    zip_buffer = exp.make_zip_binary()

    # Prepare the response, setting the HTTP headers for a downloadable file
    model_name_exported = ts_obj.__class__.__name__
    output_file = f'export_{len(datas)}_{model_name_exported}_{dt.datetime.utcnow().isoformat()}.zip'
    response = HttpResponse(zip_buffer, content_type='application/zip')
    response['Content-Disposition'] = f'attachment; filename="{output_file}"'

    return response

@admin.action(description="Download selected timeseries chunks")
def download_timeseries_from_chunkstore(modeladmin, request, queryset):
    holc_ts_qs = queryset.all()
    if not holc_ts_qs.exists():
        modeladmin.message_user(
            request,
            gettext("Please select at least one object"),
            messages.WARNING,
        )
        return

    datas = []
    ts_obj = None
    for ts_obj in holc_ts_qs:
        obj_content = vars(ts_obj)
        ds = ts_obj.__class__._decompress(ts_obj)
        obj_content.pop("data")
        obj_content.pop("id", None)
        obj_content = {k: v for k, v in obj_content.items() if k[0] != "_"}
        datas.append((ds, obj_content))

    exp = CompressedExport()
    summary_data = []
    for ii, (ds, obj_content) in enumerate(datas):
        filename = f"export_serie_{ii}.csv"
        df = ds.to_frame(name="data")
        exp.append_df_attach(df, filename)
        summary_data.append({"filename": filename, **obj_content})

    df_summary = pd.DataFrame(summary_data)
    exp.append_df_attach(df_summary, "content_summary.csv")

    zip_buffer = exp.make_zip_binary()

    model_name_exported = ts_obj.__class__.__name__
    output_file = (
        f"export_{len(datas)}_{model_name_exported}_{dt.datetime.utcnow().isoformat()}.zip"
    )
    response = HttpResponse(zip_buffer, content_type="application/zip")
    response["Content-Disposition"] = f'attachment; filename="{output_file}"'

    return response


import datetime as dt
import io
import logging
from collections import defaultdict
from typing import List, Dict

import pandas as pd
from django.db import models
from django.db.models import Max
from packaging.version import Version
from pyarrow import BufferReader

from hostore.models.manager import StoreQuerySet
from hostore.utils.range.range import Range
from hostore.utils.timeseries import ts_combine_first, check_ts_completeness, find_constant_sequences
from hostore.utils.utils import chunks, slice_with_delay

logger = logging.getLogger(__name__)
# below this version of pandas we must remove datetime index from the serie to be feathered
MIN_PANDAS_VERSION_FEATHER_SAVE_DATETIME_INDEX = '2.2.0'


class Store(models.Model):
    client_id = models.IntegerField()
    prm = models.CharField(max_length=70)
    last_modified = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True, null=True)
    data = models.BinaryField(blank=True, null=True)
    version = models.IntegerField(default=0)

    objects = StoreQuerySet.as_manager()

    class Meta:
        abstract = True
        constraints = [models.UniqueConstraint(fields=['prm', 'client_id', 'version'],
                                               name='%(app_label)s_%(class)s_prm_clientid_version'), ]
        indexes = [models.Index(fields=['prm', 'client_id']), ]

    @classmethod
    def count(cls, client_id: int, custom_filters=None):
        if custom_filters is None:
            custom_filters = {}
        return cls.objects.filter(client_id=client_id, **custom_filters).count()

    @classmethod
    def find_groups(self, client_id: int, prms: [str], drr: Range,
                    freq='30min',
                    combined_by=('prm',),
                    order_by=('-version',)) -> (dict, dict):
        """
        Groups parameters (`prms`) based on missing data ranges for a specific `client_id`
        and the requested date range (`drr`). This method identifies time periods where
        the data is missing or null for each parameter, and returns a grouping of those ranges.

        Parameters:
        ----------
        client_id : int
            The ID of the client for whom data is being retrieved.
        prms : List[str]
            A list of parameter names (strings) whose data availability is being checked.
        drr : Range
            A `Range` object representing the requested date range. Limits included []
        freq : str, optional
            The frequency used to detect missing data gaps, defaults to '30min'.
        combined_by : tuple, optional
            A tuple specifying how the data should be grouped (default is by 'prm').
        order_by : tuple, optional
            A tuple specifying how the data should be ordered (default is by '-version').

        Returns:
        -------
        groups : dict
            A dictionary where each key is a `Range` object representing a missing data
            range, and the corresponding value is a list of parameters (`prms`) that have
            missing data in that range.

        data : dict
            A dictionary containing the retrieved data for the provided parameters,
            keyed by the parameter name. Each value is a list of dictionaries representing
            the time series data for the parameter.

        Raises:
        ------
        AssertionError
            If more than one object is retrieved for a given parameter, which indicates
            a problem with the data retrieval.

        Notes:
        ------
        - The method first retrieves data using the `get_many_lc` method, combining versions
          of the data as necessary and ordering them as per the `order_by` argument.
        - Any parameters for which no data is found are considered "absent" and are grouped
          under the entire requested date range (`drr`).
        - For each parameter with available data, the method slices the data according to
          the requested date range and identifies missing data by comparing the available
          range (`dra`) to the requested range (`drr`).
        - Null sequences within the time series are also identified and treated as missing
          data gaps.
        - Missing or null data ranges are merged into larger combined ranges to avoid
          fragmentation, and these combined ranges are used to group parameters in the
          `groups` dictionary.

        Example:
        --------
        >>> groups, data = Store.find_groups(
                client_id=123,
                prms=['prm1', 'prm2'],
                drr=Range(sd='2023-01-01', ed='2023-02-01'),
                freq='30min'
            )
        >>> print(groups)
        {Range(sd='2023-01-01', ed='2023-01-10'): ['param1'], Range(sd='2023-01-15', ed='2023-01-20'): ['param2']}
        >>> print(data)
        {'param1': [...], 'param2': [...]}
        """

        data = self.get_many_lc(prms, client_id, combined_versions=True, combined_by=combined_by, order_by=order_by,)

        groups = defaultdict(lambda: [], {})
        prms_absents = set(prms) - set(data.keys())

        # All absent keys start from start_date
        if len(prms_absents) > 0:
            groups[drr].extend(prms_absents)

        keys_to_pop = []
        for k, objs in data.items():
            assert len(objs) == 1
            obj = objs[0]
            obj['data'] = obj['data'][drr.sd:drr.ed]
            mask_not_null = obj['data'].notnull()
            lc = obj['data'][mask_not_null]
            # dra : date range available
            # drr : date range requested
            missing_ranges = []
            if lc.empty:
                missing_ranges.append(drr)
                keys_to_pop.append(k)
            else:
                dra = Range(lc.index.min(), lc.index.max())
                _missing_ranges = dra.difference_missing(drr, min_delta=pd.to_timedelta(freq))
                missing_ranges.extend(_missing_ranges)

            # Find holes in not null data
            if not mask_not_null.empty:
                constant_seqs = find_constant_sequences(mask_not_null)
                for (start, end), not_null in constant_seqs:
                    if not not_null:
                        logger.info(f"key {k}, Found a null sequence between {start} - {end}")
                        missing_ranges.append(Range(start, end))

            # To prevent multiple ranges we combine them
            combined_ranges = Range.combine(missing_ranges)
            for combined_range in combined_ranges:
                groups[combined_range].append(k)

        # pop keys with empty data
        [data.pop(k) for k in keys_to_pop]

        return groups, data

    @classmethod
    def find_holes(cls, client_id: int, sd: dt.datetime, ed: dt.datetime, freq='30min', prms: List[str] = None,
                   chunk_size=50,
                   freq_margin=pd.Timedelta(minutes=0),
                   custom_filters=None,
                   combined_by=('prm',), order_by=('-version',)) -> Dict[str, List]:
        """
        Find holes (missing data periods) for multiple prms within a specified datetime range.

        This method retrieves data from a database for each parameter associated with a client_id,
        checks each dataset for missing data within the specified datetime range, and identifies the start
        and end of each missing data period.

        :param client_id: ID of the client for whom the data is being checked.
        :param sd: Start datetime of the range to check for holes.
        :param ed: End datetime of the range to check for holes.
        :param freq: Frequency at which the data is expected, defaulting to '30min'.
        :param prms: Optional list of prms to check; if None, all parameters for the client are checked.
        :param chunk_size: The size of chunks to break the parameter list into for batch processing.
                           This helps in managing memory and processing large datasets efficiently, defaulting to 50.
        :param freq_margin: Freq margin for null sequences
        :param custom_filters: (optionnal) None or dict
        :param combined_by: set of attributes for ordering load_curves before combine_first
        :return: A dictionary with prm as keys and a list of tuples (start, end) indicating
                 the missing data periods for each parameter. Start and end are included
        """
        if custom_filters is None:
            custom_filters = {}
        # If no prm are specified, retrieve all prm for the given client from the database
        if prms is None:
            prms = list(cls.objects.filter(client_id=client_id, **custom_filters).values_list('prm', flat=True))

        # Iterate over prms in chunks to manage memory usage and efficiency
        for prm_chunk in chunks(prms, chunk_size):
            data_by_prm = cls.get_many_lc(prm_chunk, client_id,
                                          combined_versions=True,
                                          combined_by=combined_by,
                                          order_by=order_by,
                                          **custom_filters)

            # Check each parameter's data for completeness
            for prm in prm_chunk:
                nulls_seqs = []
                data = data_by_prm[prm]
                if len(data) == 0:
                    # If there's no data for a prm, the entire range is considered as a hole
                    nulls_seqs = [(sd, ed)]
                else:
                    # Data is wrapped in a list, and there's only one dataset per prm (combined_versions=True)
                    assert len(data) == 1
                    nulls_seqs = check_ts_completeness(data[0]['data'], sd, ed, freq=freq, freq_margin=freq_margin)

                # Yield the output for this prm
                # This allows processing part by part without waiting for the entire operation to complete
                yield prm, nulls_seqs

    @classmethod
    def get_lc(cls, prm: str, client_id: int, combined_versions=True, version: int = None, custom_filters=None,
               combined_by=('prm',),
               order_by=('-version',),
               combined_delay=None) -> List[Dict]:
        """
        Get the prm load curve
        Args:
            prm: str
            client_id: int, client's id
            combined_versions: bool, if multiple versions exists we combine them (default True)
            version: return only the selected version
            custom_filters: (optionnal) None or dict
            combined_by: set of attributes to bombined by
            order_by: set of attributes for ordering load_curves before combine_first
            combined_delay : pd.Timedelta object, optional to slice the beginning of the time series
        Returns:
            List[Dict]
        """
        logger.info(f'Getting load curve with prm {prm} for client {client_id}')
        if custom_filters is None:
            custom_filters = {}

        qs = cls.objects.filter(prm=prm, client_id=client_id, **custom_filters).order_by(*order_by)
        if version is not None:
            qs = qs.filter(version=version)

        entries = []
        for entry in qs:
            entry_dict = vars(entry)
            reader = BufferReader(entry_dict['data'])
            ds = pd.read_feather(reader)
            # fix feather index
            if 'index' in ds.columns:
                ds.set_index('index', inplace=True)

            entry_dict['data'] = ds.iloc[:, 0]
            entries.append(entry_dict)

        cleared_combined_by = set([attr for attr in combined_by])
        ds_combined_dict = defaultdict(lambda: None)
        if len(entries) > 0 and combined_versions:
            for e in entries:
                key = str([e[attr] for attr in cleared_combined_by])
                if combined_delay is not None:
                    e['data'] = slice_with_delay(e['data'], combined_delay)
                if ds_combined_dict[key] is None:
                    ds_combined_dict[key] = e
                else:
                    ds_combined_dict[key]['data'] = ts_combine_first([ds_combined_dict[key]['data'], e['data']])
            entries = list(ds_combined_dict.values())
        return entries

    @classmethod
    def get_many_lc(cls, prms: [str], client_id: int, combined_versions=True,
                    custom_filters=None,
                    combined_by=('prm',),
                    order_by=('-version',),
                    combined_delay=None) -> Dict[str, List[Dict]]:
        """
        Get many prms from cache
        Args:
            prms: list of prms
            client_id: int, client's id
            combined_versions: bool, if multiple versions exists we combine them (default True)
            custom_filters: dict used in filter
            combined_by: set of attributes to bombined by
            order_by: set of attributes for ordering load_curves before combine_first
            combined_delay : pd.Timedelta object, optional to slice the beginning of the time series
        Returns:
            Dict[str, List[Dict]]
        """
        if custom_filters is None:
            custom_filters = {}
        qs = cls.objects.filter(prm__in=prms, client_id=client_id, **custom_filters).order_by(*order_by)
        results = defaultdict(lambda: [])
        for entry in qs:
            reader = BufferReader(entry.data)
            ds = pd.read_feather(reader)
            # fix feather index
            if 'index' in ds.columns:
                ds.set_index('index', inplace=True)
            entry_dict = vars(entry)
            entry_dict['data'] = ds.iloc[:, 0]
            results[entry.prm].append(entry_dict)

        cleared_combined_by = set([attr for attr in combined_by])
        ds_combined_dict = defaultdict(lambda: defaultdict(lambda: None))
        if len(results) > 0 and combined_versions:
            for prm, entries in results.items():
                for e in entries:
                    key = str([e[attr] for attr in cleared_combined_by])
                    if combined_delay is not None:
                        e['data'] = slice_with_delay(e['data'], combined_delay)
                    if ds_combined_dict[prm][key] is None:
                        ds_combined_dict[prm][key] = e
                    else:
                        ds_combined_dict[prm][key]['data'] = ts_combine_first([ds_combined_dict[prm][key]['data'], e['data']])
            for prm in results:
                results[prm] = list(ds_combined_dict[prm].values())

        return results

    @classmethod
    def set_lc(cls, prm: str,
               value: pd.Series,
               client_id: int,
               versionning=False,
               versionning_by=('prm',),
               attributes_to_set=None) -> None:
        """
        Set one prm to cache
        Args:
            prm: str
            value: pd.Series
            client_id: int, client's id
            versionning: bool, if True, a new version will be saved else the load curve is replaced (default False)
            versionning_by: set of attributes used for versionning (prm is mandatory)
            attributes_to_set: (optionnal) None or dict
        Returns:
            None
        """
        if attributes_to_set is None:
            attributes_to_set = {}
        logger.info(f'SET key {prm} in cache')
        if isinstance(value, pd.Series):
            if value.isnull().all():
                logger.warning(f'CACHE : Key {prm} is ignored because data is null')
                return
            df = value.to_frame(name=prm)
            # fix feather index
            if Version(pd.__version__) < Version(MIN_PANDAS_VERSION_FEATHER_SAVE_DATETIME_INDEX):
                df.reset_index(inplace=True, names=['index'])
            buf = io.BytesIO()
            df.to_feather(buf, compression='lz4')
            v = buf.getvalue()
            if not versionning:
                cls.objects.update_or_create(client_id=client_id, prm=prm, defaults=dict(data=v), **attributes_to_set)
            else:
                assert 'prm' in versionning_by
                _filters = {'prm': prm, **{k: v for k, v in attributes_to_set.items() if k in versionning_by}}
                latest_version = cls.objects.filter(client_id=client_id, **_filters).aggregate(Max('version'))['version__max']
                if latest_version is not None:
                    # If there's at least one object with the same prm, increment the version
                    version = latest_version + 1
                else:
                    # If there are no objects with the same prm, start with version 0
                    version = 0
                logger.info(f"version to insert {version}")
                cls.objects.create(client_id=client_id, prm=prm, data=v, version=version, **attributes_to_set)
            logger.info(f'SET prm {prm} in cache DONE')
        else:
            raise ValueError(f'Cannot cache value of type {type(value)}, only dataframe are accepted')

    @classmethod
    def set_many_lc(cls, dataseries: Dict[str, pd.Series], client_id: int, versionning=False,
                    versionning_by=('prm',),
                    attributes_to_set=None) -> None:
        """
        Update prms contained in dataseries dict
        Args:
            dataseries: dict(key: pd.Series)
            client_id: int, client's id
            versionning: bool, if True, a new version will be saved for each prm else the load curve is replaced
            versionning_by: set of attributes used for versionning (prm is mandatory)
            attributes_to_set: (optionnal) None or dict
        Returns:
            None
        """
        if len(dataseries) > 0:
            logger.info(f'Updating cache for {list(dataseries.keys())}')
        for prm, data in dataseries.items():
            if data is not None and not data.empty:
                cls.set_lc(prm, data, client_id,
                           versionning=versionning,
                           versionning_by=versionning_by,
                           attributes_to_set=attributes_to_set)

    @classmethod
    def clear(cls, prms: [str], client_id: int, version=None, custom_filters=None) -> None:
        """
        Method to clear multiple prm from cache
        Args:
            prms: list of prms
            client_id: client's id
            version: version id to delete
            custom_filters: (optionnal) None or dict
        Returns:
            None
        """
        if custom_filters is None:
            custom_filters = {}
        qs = cls.objects.filter(prm__in=prms, client_id=client_id, **custom_filters)
        if version is not None:
            qs = qs.filter(version=version)
        qs.delete()

    @classmethod
    def clear_all(cls, client_id: int = None, custom_filters=None) -> None:
        """
        Method to clear multiple prm from cache
        Args:
            client_id: client's id
            custom_filters: (optionnal) None or dict
        Returns:
            None
        """
        if custom_filters is None:
            qs = cls.objects.all()
        else:
            qs = cls.objects.filter(**custom_filters)
        if client_id is not None:
            qs.filter(client_id=client_id).delete()
        else:
            qs.delete()




# ts_sync/views.py
import functools
import logging
import time
import traceback

import backoff
import requests

from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.pagination import LimitOffsetPagination
from rest_framework.response import Response
import base64, pandas as pd
from typing import Type
import datetime as dt

DEBUG = False


logger = logging.getLogger(__name__)

def print_api_exception(view_func):
    """
    affiche erreur api dans une view
    """
    @functools.wraps(view_func)
    def wrapper(self, request, *args, **kw):
        try:
            return view_func(self, request, *args, **kw)
        except Exception as exc:
            if DEBUG:
                print(f'\nview func {view_func} unexpected error')
                print(traceback.format_exc())
            raise
    return wrapper


class ChunkIteratorPagination(LimitOffsetPagination):
    """Limit/offset paginator that streams queryset results."""

    def paginate_queryset(self, queryset, request, view=None):
        self.limit = self.get_limit(request)
        if self.limit is None:
            return None

        self.count = self.get_count(queryset)
        self.request = request
        self.offset = self.get_offset(request)

        if self.count == 0 or self.offset > self.count:
            return []

        chunk_size = getattr(view, "qs_iterator_chunk_size", 200)
        return list(
            queryset[self.offset : self.offset + self.limit]
        )


class TimeseriesChunkStoreSyncViewSet(viewsets.GenericViewSet):
    """
    Base server-side ViewSet that exposes a REST interface for synchronising
    any TimeseriesChunkStore subclass.

    End-points
    ──────────
    • GET  /updates/?since=ISO    → list of modified chunks
    • GET /pack/                 → export requested chunks

    Usage example in your own views.py:

        class YearStoreSyncView(TimeseriesChunkStoreSyncViewSet):
            store_model = TestStoreChunkYear     # ← your store

        router.register("ts/year", YearStoreSyncView, basename="ts-year")
    """

    store_model: Type['TimeseriesChunkStore'] = None
    pagination_class = ChunkIteratorPagination

    # 1) /updates/?since=ISO
    @action(detail=False, methods=["get"])
    @print_api_exception
    def updates(self, request):
        if not self.store_model.ALLOW_CLIENT_SERVER_SYNC:
            raise ValueError(f'Trying to use TimeseriesChunkStoreSyncViewSet with model {self.store_model.__name__} '
                             f'while ALLOW_CLIENT_SERVER_SYNC=False.')

        since = pd.Timestamp(request.query_params["since"])
        filters = {
            k: v
            for k, v in request.query_params.items()
            if k not in {"since", "limit", "offset"}
        }

        # log query
        try:
            user = self.request.user
        except:
            user = None
        logger.info(f'Sync updates for model {self.store_model.__name__} required by user {user} since {since} (filters {filters})')

        qs = self.store_model.updates_queryset(since, filters)
        page = self.paginate_queryset(qs)
        data = [
            {
                "attrs": {k: row[k] for k in self.store_model.get_model_keys()},
                "chunk_index": row["chunk_index"],
                "dtype": row["dtype"],
                "start_ts": row["start_ts"],
                "updated_at": row["updated_at"],
                "is_deleted": row["is_deleted"],
            }
            for row in page
        ]
        return self.get_paginated_response(data)

    # 2) /pack/   GET → export
    @action(detail=False, methods=["get"])
    @print_api_exception
    def pack(self, request):
        if not self.store_model.ALLOW_CLIENT_SERVER_SYNC:
            raise ValueError(f'Trying to use TimeseriesChunkStoreSyncViewSet with model {self.store_model.__name__} '
                             f'while ALLOW_CLIENT_SERVER_SYNC=False.')

        spec    = request.data

        # log query
        try:
            user = self.request.user
        except:
            user = None
        logger.info(
            f'Sync pack for model {self.store_model.__name__} required by user ({len(spec)} pieces)')

        chunks  = self.store_model.export_chunks(spec)
        payload = [
            {
                "blob":  base64.b64encode(b).decode(),
                "attrs": attrs,
                "meta":  meta,
            } for b, attrs, meta in chunks
        ]
        return Response(payload, content_type="application/json")

    @classmethod
    def as_factory(cls, model, **extra_attrs):
        """
        Dynamically create a ViewSet bound to *model* and
        enriched with optional DRF attributes.

        Example
        -------
        YearSync = TimeseriesChunkStoreSyncViewSet.as_factory(
            TestStoreChunkYear,
            permission_classes=[IsAuthenticatedActive],
            parser_classes=[JSONParser],
            throttle_classes=[],               # disable throttling
        )
        router.register("ts/year", YearSync, basename="ts-year")
        """
        if not hasattr(model, "updates_queryset"):
            raise TypeError("model must inherit TimeseriesChunkStore")
        attrs = {"store_model": model, **extra_attrs}
        return type(f"{model.__name__}SyncViewSet", (cls,), attrs)


class TimeseriesChunkStoreSyncClient:
    """
    client = TimeseriesChunkStoreSyncClient(
                endpoint="https://api.example.com/ts/year/",
                store_model=TestStoreChunkYear)
    client.pull(since) : récupère les nouveautés
    """

    def __init__(self, *,
                 endpoint: str,
                 store_model: Type['TimeseriesChunkStore'],
                 retry_max_tries: int = 5,
                 retry_max_time: int = 300,
                 requests_get_kwargs=None
                 ):
        """
        Construct the synchronization client for store_model.

        Args:
            endpoint: url to call /updates/ and /pack/
            store_model: client TimeseriesChunkStore model to insert updates
            retry_max_tries: number of retry attemps
            retry_max_time: retry time in milliseconds
            requests_get_kwargs: kwargs to pass to requests.get
        """
        if not store_model.ALLOW_CLIENT_SERVER_SYNC:
            raise ValueError(f'Trying to use TimeseriesChunkStoreSyncClient with model {store_model.__name__} '
                             f'while ALLOW_CLIENT_SERVER_SYNC=False.')

        self.endpoint = endpoint.rstrip("/")
        self.store_model    = store_model
        self._retry_tries = retry_max_tries
        self._retry_time  = retry_max_time
        self.requests_get_kwargs  = requests_get_kwargs if requests_get_kwargs is not None else {}

    # ----------- pull depuis le serveur -------------------------------
    def pull(
        self,
        filters: dict | None = None,
        batch: int = 50,
        page_size: int = 200,
    ):
        """Fetch updates from the server in a paginated fashion.

        Args:
            filters: optional server-side filters.
            batch: chunk size for `/pack/` requests.
            page_size: number of items to request from `/updates/` per call.
        """
        filters = dict(filters or {})

        since = self.store_model.last_updated_at(filters)
        url_updates = f"{self.endpoint}/updates/"
        params = {"since": since.isoformat(), "limit": page_size, **filters}
        total_fetch = total_delete = 0

        def _parse_updated_at(value):
            ts = pd.Timestamp(value)
            if ts.tzinfo is None:
                ts = ts.tz_localize("UTC")
            return ts.to_pydatetime()

        while url_updates:
            page = self._get(url_updates, params=params)
            params = None  # next links already contain query parameters
            updates = page.get("results", [])
            if not updates:
                break

            to_fetch, to_delete = [], []
            for u in updates:
                (to_delete if u["is_deleted"] else to_fetch).append(u)

            for d in to_delete:
                qs = self.store_model.objects.filter(
                    **d["attrs"], chunk_index=d["chunk_index"]
                )
                qs.delete(keep_tracking=True)
                qs.update(updated_at=_parse_updated_at(d["updated_at"]))

            for i in range(0, len(to_fetch), batch):
                spec = to_fetch[i : i + batch]
                pack = self._get(f"{self.endpoint}/pack/", json=spec)
                tuples = [
                    (base64.b64decode(item["blob"]), item["attrs"], item["meta"])
                    for item in pack
                ]
                self.store_model.import_chunks(tuples)

            total_fetch += len(to_fetch)
            total_delete += len(to_delete)
            url_updates = page.get("next")

        return total_fetch, total_delete


    def _get(self, url: str, **kwargs):

        # Retry tant que _call() retourne None
        @backoff.on_predicate(
            backoff.expo,
            max_tries=self._retry_tries,
            max_time=self._retry_time,
        )
        def _call():
            resp = requests.get(url, **kwargs, **self.requests_get_kwargs)

            # Si ce n'est pas un 429 -> comportement "normal"
            if resp.status_code != 429:
                resp.raise_for_status()
                return resp.json()

            # On récupère la valeur du retry-after
            retry_after = resp.headers.get("Retry-After")

            if retry_after:
                try:
                    delay = int(retry_after)
                    logger.info(f"Rate limit reached. Waiting {delay}s (Retry-After)")
                    time.sleep(delay)
                except ValueError as e:
                    logger.error(f"Something went wrong while waiting : {retry_after} (Retry-After)")
                    # DRF ne renvoie normalement QUE des entiers
                    raise e

            # Retourne None déclenche un retry SANS compter un échec
            return None

        return _call()




class Range:

    def __init__(self, sd, ed):
        assert sd <= ed
        self.sd = sd
        self.ed = ed

    def __hash__(self):
        return hash(f'{self.sd}-{self.ed}')

    def __str__(self):
        return f'Interval from {self.sd} to {self.ed}'

    def __eq__(self, other):
        return (self.sd == other.sd) and (self.ed == other.ed)

    def overlap(self, interval):
        return (self.sd <= interval.ed) and (self.ed >= interval.sd)

    def check_date(self, date):
        """
        Check if date is in range
        """
        return (date <= self.ed) and (date >= self.sd)

    @staticmethod
    def sort_by_start(ranges, reverse=False):
        ranges.sort(key=lambda r: r.sd, reverse=reverse)

    @staticmethod
    def combine(ranges):
        Range.sort_by_start(ranges)
        itr = iter(ranges)
        current_range = next(itr, None)
        combined_ranges = []
        if current_range is not None:
            combined_ranges.append(current_range)
            next_range = next(itr, None)
            while next_range is not None:
                if combined_ranges[-1].overlap(next_range):
                    combined_ranges[-1].ed = next_range.ed
                else:
                    combined_ranges.append(next_range)
                next_range = next(itr, None)
        return combined_ranges

    def includes(self, interval):
        """
        Check if it is included in range
        """
        return self.check_date(interval.sd) and self.check_date(interval.ed)

    def intersection(self, interval, min_delta):
        start = max(interval.sd, self.sd)
        end = min(interval.ed, self.ed)
        if end - start >= min_delta:
            return Range(start, end)

    def difference_missing(self, interval, min_delta):
        result = []
        if self.overlap(interval):
            if self.includes(interval):
                return result
            elif interval.includes(self):
                if (self.sd - interval.sd) >= min_delta:
                    result.append(Range(interval.sd, self.sd))
                if (interval.ed - self.ed) >= min_delta:
                    result.append(Range(self.ed, interval.ed))
                return result
            else:
                pass
        if interval.ed > self.ed:
            if (interval.ed - self.ed) >= min_delta:
                return [Range(self.ed, interval.ed)]
        elif interval.sd < self.sd:
            if (self.sd - interval.sd) >= min_delta:
                return [Range(interval.sd, self.sd)]
        else:
            pass
        return result



import io
import zipfile

import pandas as pd
from unittest.mock import Mock

from django.contrib import admin
from django.contrib.admin import AdminSite
from django.contrib.auth import get_user_model
from django.db import models
from django.http import HttpResponse
from django.test import TransactionTestCase

from hostore.admin_actions import download_timeseries_from_store, download_timeseries_from_chunkstore
from hostore.models import TimeseriesStore, TimeseriesChunkStore
from hostore.utils.utils_test import TempTestTableHelper


def gen_serie(start, end, data, freq='1h'):
    dt_rng = pd.date_range(start, end, freq=freq)
    return pd.Series(data, index=dt_rng, name='data')


class TestAdminTimeseriesStoreWithAttribute(TimeseriesStore):
    year = models.IntegerField()
    kind = models.CharField(max_length=100)

    class Meta(TimeseriesStore.Meta):
        abstract = False
        constraints = [models.UniqueConstraint(fields=['year', 'kind'], name='hostore_TestAdminTimeseriesStoreWithAttribute_unq'), ]
        app_label = 'ts_inline'
        managed = True


class TestTimeseriesStoreWithAttributeAdmin(admin.ModelAdmin):
    resource_classes = [TestAdminTimeseriesStoreWithAttribute]
    list_display = ('year', 'kind', )
    list_filter = ('year', 'kind', )
    actions = [download_timeseries_from_store]


class DownloadTimeseriesAdminActionTest(TransactionTestCase, TempTestTableHelper):
    databases = ('default',)
    test_table = TestAdminTimeseriesStoreWithAttribute

    def setUp(self):
        self._ensure_tables()
        # populate TestTimeseriesStoreWithAttribute
        ts_attrs_y_2020_kind_a = dict(year=2020, kind='a')
        ds_y_2020_kind_a = gen_serie("2020-01-01 00:00:00+00:00", "2020-01-01 02:00:00+00:00", [1, 2, 3])
        TestAdminTimeseriesStoreWithAttribute.set_ts(ts_attrs_y_2020_kind_a, ds_y_2020_kind_a)
        self.ts_attrs_y_2020_kind_a = ts_attrs_y_2020_kind_a
        self.ds_y_2020_kind_a = ds_y_2020_kind_a

        ts_attrs_y_2020_kind_b = dict(year=2020, kind='b')
        ds_y_2020_kind_b = gen_serie("2020-01-01 00:00:00+01:00", "2020-01-01 02:00:00+01:00", [10, 20, 30])
        TestAdminTimeseriesStoreWithAttribute.set_ts(ts_attrs_y_2020_kind_b, ds_y_2020_kind_b)
        self.ts_attrs_y_2020_kind_b = ts_attrs_y_2020_kind_b
        self.ds_y_2020_kind_b = ds_y_2020_kind_b

        # Mock request
        self.user = get_user_model().objects.create(
            email='user@example.com', password='password', is_superuser=False)
        self.mock_request = Mock(user=self.user)

        # Create an instance of the admin site
        self.site = AdminSite()

        # Create an instance of the admin class for your model
        self.model_admin = TestTimeseriesStoreWithAttributeAdmin(
            TestAdminTimeseriesStoreWithAttribute, self.site)

    def test_zip_content(self):
        queryset = TestAdminTimeseriesStoreWithAttribute.objects.all()
        response = download_timeseries_from_store(self.model_admin, self.mock_request, queryset)
        self.assertIsInstance(response, HttpResponse)
        byte_content = response.content
        zip_bytes_io = io.BytesIO(byte_content)
        with zipfile.ZipFile(zip_bytes_io, 'r') as zip_file:
            # List all files inside the zip file
            file_list = zip_file.namelist()
            self.assertListEqual(file_list, ['export_serie_0.csv', 'export_serie_1.csv', 'content_summary.csv'])
            with zip_file.open('content_summary.csv') as specific_file:
                # Read the file content (assuming it's a text file, like CSV)
                file_content = specific_file.read().decode('utf-8')
                expected = ';filename;year;kind\n0;export_serie_0.csv;2020;a\n1;export_serie_1.csv;2020;b\n'
                self.assertEquals(file_content, expected)


class TestAdminTimeseriesChunkStore(TimeseriesChunkStore):
    version = models.IntegerField()
    kind = models.CharField(max_length=100)
    STORE_TZ = "UTC"

    class Meta(TimeseriesChunkStore.Meta):
        app_label = "ts_inline"
        managed = True


class TestTimeseriesChunkStoreAdmin(admin.ModelAdmin):
    resource_classes = [TestAdminTimeseriesChunkStore]
    actions = [download_timeseries_from_chunkstore]


class DownloadTimeseriesChunkAdminActionTest(TransactionTestCase, TempTestTableHelper):
    databases = ("default",)
    test_table = TestAdminTimeseriesChunkStore

    def setUp(self):
        self._ensure_tables()
        attrs_a = dict(version=1, kind="a")
        ds_a = gen_serie(
            "2020-01-01 00:00:00+00:00",
            "2020-01-01 02:00:00+00:00",
            [1, 2, 3],
        )
        TestAdminTimeseriesChunkStore.set_ts(attrs_a, ds_a)

        attrs_b = dict(version=1, kind="b")
        ds_b = gen_serie(
            "2020-01-01 00:00:00+00:00",
            "2020-01-01 02:00:00+00:00",
            [10, 20, 30],
        )
        TestAdminTimeseriesChunkStore.set_ts(attrs_b, ds_b)

        self.user = get_user_model().objects.create(
            email="user@example.com", password="password", is_superuser=False
        )
        self.mock_request = Mock(user=self.user)

        self.site = AdminSite()
        self.model_admin = TestTimeseriesChunkStoreAdmin(
            TestAdminTimeseriesChunkStore, self.site
        )

    def test_zip_content(self):
        queryset = TestAdminTimeseriesChunkStore.objects.all()
        response = download_timeseries_from_chunkstore(
            self.model_admin, self.mock_request, queryset
        )
        self.assertIsInstance(response, HttpResponse)
        byte_content = response.content
        zip_bytes_io = io.BytesIO(byte_content)
        with zipfile.ZipFile(zip_bytes_io, "r") as zip_file:
            file_list = zip_file.namelist()
            self.assertListEqual(
                file_list,
                ["export_serie_0.csv", "export_serie_1.csv", "content_summary.csv"],
            )
            with zip_file.open("content_summary.csv") as specific_file:
                file_content = specific_file.read().decode("utf-8")
                self.assertIn("version", file_content)
                self.assertIn("kind", file_content)


from django.db import connection


class TempTestTableHelper:
    test_table = None

    def _ensure_tables(self, delete_kw=None):
        """(Re)crée les tables manquantes après un flush Django."""
        delete_kw = delete_kw or dict()
        existing = connection.introspection.table_names()
        if self.test_table is None:
            raise ValueError('test_table is None')
        with connection.schema_editor(atomic=False) as se:
            if self.test_table._meta.db_table not in existing:
                se.create_model(self.test_table)
        self.test_table.objects.all().delete(**delete_kw)



import pandas as pd
from django.db import models
from django.test import TestCase, TransactionTestCase

from hostore.models import TimeseriesStore
from hostore.utils.timeseries import ts_combine_first
from hostore.utils.utils_test import TempTestTableHelper


# ./manage.py test hostore.tests.tests_timeseries_store.TimeseriesCacheTestCase

def gen_serie(start, end, data, freq='1h'):
    dt_rng = pd.date_range(start, end, freq=freq)
    return pd.Series(data, index=dt_rng, name='data')


class TestTimeseriesStoreWithAttribute(TimeseriesStore):
    year = models.IntegerField()
    kind = models.CharField(max_length=100)

    class Meta(TimeseriesStore.Meta):
        abstract = False
        constraints = [models.UniqueConstraint(fields=['year', 'kind'], name='hostore_TestTimeseriesStoreWithAttribute_unq'), ]
        app_label = 'ts_inline'
        managed = True

class TimeseriesCacheTestCase(TransactionTestCase, TempTestTableHelper):
    databases = ('default',)
    test_table = TestTimeseriesStoreWithAttribute

    def setUp(self):
        self._ensure_tables()
        ts_attrs_y_2020_kind_a = dict(year=2020, kind='a')
        ds_y_2020_kind_a = gen_serie("2020-01-01 00:00:00+00:00", "2020-01-01 02:00:00+00:00", [1, 2, 3])
        TestTimeseriesStoreWithAttribute.set_ts(ts_attrs_y_2020_kind_a, ds_y_2020_kind_a)
        self.ts_attrs_y_2020_kind_a = ts_attrs_y_2020_kind_a
        self.ds_y_2020_kind_a = ds_y_2020_kind_a

        ts_attrs_y_2020_kind_b = dict(year=2020, kind='b')
        ds_y_2020_kind_b = gen_serie("2020-01-01 00:00:00+01:00", "2020-01-01 02:00:00+01:00", [10, 20, 30])
        TestTimeseriesStoreWithAttribute.set_ts(ts_attrs_y_2020_kind_b, ds_y_2020_kind_b)
        self.ts_attrs_y_2020_kind_b = ts_attrs_y_2020_kind_b
        self.ds_y_2020_kind_b = ds_y_2020_kind_b
        
        ts_attrs_y_2024_kind_a = dict(year=2024, kind='a')
        ds_y_2024_kind_a = gen_serie("2024-01-01 00:00:00+00:00", "2024-01-01 02:00:00+00:00", [11, 21, 31])
        TestTimeseriesStoreWithAttribute.set_ts(ts_attrs_y_2024_kind_a, ds_y_2024_kind_a)
        self.ts_attrs_y_2024_kind_a = ts_attrs_y_2024_kind_a
        self.ds_y_2024_kind_a = ds_y_2024_kind_a

    def test_get_w_attributes(self):

        # test get - 2020 a
        data = TestTimeseriesStoreWithAttribute.get_ts(self.ts_attrs_y_2020_kind_a)
        pd.testing.assert_series_equal(data[0]['data'], self.ds_y_2020_kind_a, check_names=False, check_freq=False)

        # test get - 2020 a - flat
        ds = TestTimeseriesStoreWithAttribute.get_ts(self.ts_attrs_y_2020_kind_a, flat=True)
        pd.testing.assert_series_equal(ds, self.ds_y_2020_kind_a, check_names=False, check_freq=False)

        # test get - 2020 b (tz +1)
        data = TestTimeseriesStoreWithAttribute.get_ts(self.ts_attrs_y_2020_kind_b)
        ds = data[0]['data']
        ds = ds
        # we need to tz_convert, otherwise comparison fails :
        # Attribute "dtype" are different
        # [left]:  datetime64[ns, pytz.FixedOffset(60)]
        # [right]: datetime64[ns, UTC+01:00]
        ds = ds.tz_convert('utc')
        ds_y_2020_kind_b = self.ds_y_2020_kind_b.tz_convert('utc')
        pd.testing.assert_series_equal(
           ds, ds_y_2020_kind_b, check_names=False, check_freq=False)

        # test get - 2024 a
        data = TestTimeseriesStoreWithAttribute.get_ts(self.ts_attrs_y_2024_kind_a)
        pd.testing.assert_series_equal(data[0]['data'], self.ds_y_2024_kind_a, check_names=False, check_freq=False)

    def test_get_w_attributes_multiple(self):
        # check we found the 2 series registered as kind a
        data = TestTimeseriesStoreWithAttribute.get_ts(dict(kind='a'))
        assert len(data) == 2
        assert data[0]['year'] == 2020
        assert data[1]['year'] == 2024

        # check flat option fails (multiple series for a request)
        with self.assertRaises(ValueError):
            data = TestTimeseriesStoreWithAttribute.get_ts(dict(kind='a'), flat=True)

    def test_get_w_nodata(self):
        # check we found the 2 series registered as kind a
        data = TestTimeseriesStoreWithAttribute.get_ts(dict(kind='z'))
        assert len(data) == 0

        # check flat option fails (no series for a request)
        with self.assertRaises(ValueError):
            data = TestTimeseriesStoreWithAttribute.get_ts(dict(kind='z'), flat=True)

    def test_set_wout_update(self):
        ts_attrs_y_2010_kind_wout_upd = dict(year=2010, kind='wout_upd')

        # set 1 2 3
        ds_y_2010_kind_wout_upd0 = gen_serie("2010-01-01 00:00:00+00:00", "2010-01-01 02:00:00+00:00", [1, 2, 3])
        TestTimeseriesStoreWithAttribute.set_ts(ts_attrs_y_2010_kind_wout_upd, ds_y_2010_kind_wout_upd0)

        # set 10 20 30
        ds_y_2010_kind_wout_upd1 = gen_serie("2010-01-01 03:00:00+00:00", "2010-01-01 05:00:00+00:00", [10, 20, 30])
        with self.assertRaises(ValueError):
            # forbidden to set another series over existing attribute without update option
            TestTimeseriesStoreWithAttribute.set_ts(ts_attrs_y_2010_kind_wout_upd, ds_y_2010_kind_wout_upd1)

        # test we get first set data
        data = TestTimeseriesStoreWithAttribute.get_ts(ts_attrs_y_2010_kind_wout_upd)
        pd.testing.assert_series_equal(data[0]['data'], ds_y_2010_kind_wout_upd0, check_names=False, check_freq=False)

    def test_set_with_replace(self):
        ts_attrs_y_2010_kind_with_replace = dict(year=2010, kind='with_replace')

        # set init
        ds_y_2010_kind_with_replace0 = gen_serie("2010-01-01 00:00:00+00:00", "2010-01-01 02:00:00+00:00", [1, 2, 3])
        TestTimeseriesStoreWithAttribute.set_ts(ts_attrs_y_2010_kind_with_replace, ds_y_2010_kind_with_replace0)

        # set replace
        ds_y_2010_kind_with_replace1 = gen_serie("2010-01-01 02:00:00+00:00", "2010-01-01 05:00:00+00:00", [999, 10, 20, 30])
        TestTimeseriesStoreWithAttribute.set_ts(ts_attrs_y_2010_kind_with_replace, ds_y_2010_kind_with_replace1, replace=True)

        # test we get last data
        ds = TestTimeseriesStoreWithAttribute.get_ts(ts_attrs_y_2010_kind_with_replace, flat=True)
        pd.testing.assert_series_equal(ds, ds_y_2010_kind_with_replace1, check_names=False, check_freq=False)

    def test_set_with_update(self):
        ts_attrs_y_2010_kind_with_upd = dict(year=2010, kind='with_upd')

        # set init
        ds_y_2010_kind_with_upd0 = gen_serie("2010-01-01 00:00:00+00:00", "2010-01-01 02:00:00+00:00", [1, 2, 3])
        TestTimeseriesStoreWithAttribute.set_ts(ts_attrs_y_2010_kind_with_upd, ds_y_2010_kind_with_upd0)

        # set update
        ds_y_2010_kind_with_upd1 = gen_serie("2010-01-01 02:00:00+00:00", "2010-01-01 05:00:00+00:00", [999, 10, 20, 30])
        TestTimeseriesStoreWithAttribute.set_ts(ts_attrs_y_2010_kind_with_upd, ds_y_2010_kind_with_upd1, update=True)

        # test we get combined data (the last ts set will override existing values if any
        # ds_y_2010_kind_with_upd_combined = ts_combine_first([ds_y_2010_kind_with_upd0, ds_y_2010_kind_with_upd1])
        ds_y_2010_kind_with_upd_combined = ts_combine_first([ds_y_2010_kind_with_upd1, ds_y_2010_kind_with_upd0])

        data = TestTimeseriesStoreWithAttribute.get_ts(ts_attrs_y_2010_kind_with_upd)
        pd.testing.assert_series_equal(data[0]['data'], ds_y_2010_kind_with_upd_combined, check_names=False, check_freq=False)

        ds = TestTimeseriesStoreWithAttribute.get_ts(ts_attrs_y_2010_kind_with_upd, flat=True)
        pd.testing.assert_series_equal(ds, ds_y_2010_kind_with_upd_combined, check_names=False, check_freq=False)



# Yield successive n-sized
# chunks from l.
import pandas as pd


def chunks(l, n):
    # looping till length l
    for i in range(0, len(l), n):
        yield l[i:i + n]

def slice_with_delay(ds: pd.Series, delay:pd.Timedelta):
    if ds is not None and not ds.empty:
        return ds.copy()[ds.index.min() + delay:]
    else:
        return ds



import datetime as dt

import pandas as pd
import pytz
from django.test import TransactionTestCase

from hostore.tests.tests_basics import TestDataStore
from hostore.utils.range.range import Range
from hostore.utils.utils_test import TempTestTableHelper


class HoCacheFindGroupsTestCase(TransactionTestCase, TempTestTableHelper):
    databases = ('default',)
    test_table = TestDataStore

    def setUp(self):
        self._ensure_tables()
        self.test_client_id = 1
        self.test_data = {}

        for sd, ed, prm, val in [("2024-01-01 00:00:00+00:00", "2025-01-01 00:00:00+00:00", 'prm1', 1),
                                 ("2024-01-01 00:00:00+00:00", "2025-01-01 00:00:00+00:00", 'prm2', 2),
                                 ("2023-01-01 00:00:00+00:00", "2025-01-01 00:00:00+00:00", 'prm3', 3),]:
            idx = pd.date_range(sd, ed, freq='h')
            self.test_data[prm] = pd.Series(val, index=idx, name='data')

            TestDataStore.set_lc(prm=prm, value=self.test_data[prm], client_id=self.test_client_id,
                                 versionning=False)

    def test_find_groups(self):
        # Sample input values
        prms = ['prm1', 'prm2', 'prm3']
        sd = dt.datetime(2023, 1, 1).astimezone(tz=pytz.UTC)
        ed = dt.datetime(2023, 2, 1).astimezone(tz=pytz.UTC)
        drr = Range(sd, ed)  # Mock Range instance
        freq = '1h'

        # Call the function
        groups, data = TestDataStore.find_groups(self.test_client_id, prms, drr, freq)

        # Assertions for the groups
        self.assertIn(drr, groups)
        self.assertIn('prm2', groups[drr])  # prm2 has missing data
        self.assertIn('prm1', groups[drr])  # prm1 has missing data
        self.assertNotIn('prm3', groups[drr])  # prm3 was absent

        # Assertions for the data
        self.assertNotIn('prm1', data)
        self.assertNotIn('prm2', data)
        self.assertIn('prm3', data)

        pd.testing.assert_series_equal(data['prm3'][0]['data'], self.test_data['prm3'][drr.sd:drr.ed],
                                       check_dtype=False, check_freq=False, check_names=False)









import pandas as pd
from django.test import TransactionTestCase

from hostore.tests.tests_basics import TestDataStore
from hostore.utils.timeseries import ts_combine_first
from hostore.utils.utils_test import TempTestTableHelper


class HoCacheSliceDelayTestCase(TransactionTestCase, TempTestTableHelper):
    databases = ('default',)
    test_table = TestDataStore

    def setUp(self):
        self._ensure_tables()
        # Set up data for the tests
        # For example, create a HoCache instance
        self.test_prm = 'test_prm'
        self.test_prm_2 = 'test_prm_2'
        self.test_client_id = 1

        idx = pd.date_range("2024-01-01 00:00:00+00:00", "2025-01-01 00:00:00+00:00", freq='h')
        self.test_data = pd.Series(1, index=idx, name='data')
        idx_2 = pd.date_range("2024-01-02 00:00:00+00:00", "2025-01-01 00:00:00+00:00", freq='h')
        self.test_data_2 = pd.Series(2, index=idx_2, name='data')

        TestDataStore.set_lc(prm=self.test_prm, value=self.test_data, client_id=self.test_client_id,
                             versionning=True)
        TestDataStore.set_lc(prm=self.test_prm, value=self.test_data_2, client_id=self.test_client_id,
                             versionning=True)

    def test_slice_delay(self):
        data = TestDataStore.get_lc(self.test_prm, self.test_client_id, combined_versions=True,
                                    combined_delay=pd.Timedelta(days=1))

        self.assertEquals(len(data), 1)
        # Nous sommes décalés de 1 jour
        expected_data = ts_combine_first([self.test_data_2.iloc[24:], self.test_data.iloc[24:]])
        pd.testing.assert_series_equal(data[0]['data'], expected_data, check_names=False)

        data = TestDataStore.get_lc(self.test_prm, self.test_client_id, combined_versions=True,
                                    combined_delay=pd.Timedelta(days=2))

        self.assertEquals(len(data), 1)
        # Nous sommes décalés de 2 jours
        expected_data = ts_combine_first([self.test_data_2.iloc[48:], self.test_data.iloc[48:]])
        pd.testing.assert_series_equal(data[0]['data'], expected_data, check_names=False)

        data = TestDataStore.get_many_lc([self.test_prm], self.test_client_id, combined_versions=True,
                                         combined_delay=pd.Timedelta(days=2))

        self.assertEquals(len(data), 1)
        # Nous sommes décalés de 2 jours
        expected_data = ts_combine_first([self.test_data_2.iloc[48:], self.test_data.iloc[48:]])
        pd.testing.assert_series_equal(data[self.test_prm][0]['data'], expected_data, check_names=False)







import pandas as pd
from django.db import models
from django.test import TransactionTestCase

from hostore.models import TimeseriesChunkStore
from hostore.utils.utils_test import TempTestTableHelper


class ImportStore(TimeseriesChunkStore):
    version = models.IntegerField()
    kind = models.CharField(max_length=5)
    CHUNK_AXIS = ('year', 'month')
    ALLOW_CLIENT_SERVER_SYNC = True

    class Meta:
        app_label = 'ts_inline'
        managed = True


class ImportChunksAtomicTest(TransactionTestCase, TempTestTableHelper):
    test_table = ImportStore

    def setUp(self):
        self._ensure_tables(delete_kw=dict(_disable_sync_safety=True))

    @staticmethod
    def _make_series():
        idx = pd.date_range('2024-01-01', periods=40 * 24, freq='1h', tz='UTC')
        return pd.Series(range(len(idx)), index=idx)

    def test_failed_import_rolls_back(self):
        serie = self._make_series()
        attrs = {'version': 1, 'kind': 'A'}
        self.test_table.set_ts(attrs, serie, replace=True)
        spec = [
            {'attrs': {'version': row.version, 'kind': row.kind}, 'chunk_index': row.chunk_index}
            for row in self.test_table.objects.all()
        ]
        payload = self.test_table.export_chunks(spec)
        self.test_table.objects.all().delete(_disable_sync_safety=True)

        bad_attrs = payload[1][1].copy()
        bad_attrs.pop('chunk_index')
        bad_payload = [payload[0], (payload[1][0], bad_attrs, payload[1][2])]

        with self.assertRaises(ValueError):
            self.test_table.import_chunks(bad_payload)
        self.assertEqual(self.test_table.objects.count(), 0)

        self.test_table.import_chunks(payload)
        self.assertEqual(self.test_table.objects.count(), len(payload))



import datetime as dt
from zoneinfo import ZoneInfo

import pandas as pd
from django.db import models
from django.test import TransactionTestCase

from hostore.models import Store
from hostore.utils.utils_test import TempTestTableHelper


class TestDataStoreWithAttribute(Store):
    year = models.IntegerField()
    class Meta(Store.Meta):
        abstract = False
        constraints = [models.UniqueConstraint(fields=['prm', 'client_id', 'year', 'created_at'], name='hostore_TestDataStoreWithAttribute_unq'), ]
        app_label = 'ts_inline'
        managed = True


class HoCacheWithAttributesTestCase(TransactionTestCase, TempTestTableHelper):
    databases = ('default',)
    test_table = TestDataStoreWithAttribute

    def setUp(self):
        self._ensure_tables()
        # Set up data for the tests
        # For example, create a HoCache instance
        self.test_prm = 'test_prm'
        self.test_prm_2 = 'test_prm_2'
        self.test_client_id = 1
        self.test_data = pd.Series([1, 2, 3], index=[0, 1, 2], name='data')
        self.test_data_2 = self.test_data * 2

    def test_combined_by(self):
        TestDataStoreWithAttribute.clear_all(self.test_client_id)
        data_name = 'data_name'
        created_at_one = dt.datetime.now().astimezone(ZoneInfo("Europe/Paris"))
        TestDataStoreWithAttribute.set_lc(prm=data_name, value=self.test_data, client_id=self.test_client_id,
                                          attributes_to_set=dict(created_at=created_at_one,
                                                                 last_modified=created_at_one,
                                                                 year=2024),
                                          versionning=True,
                                          versionning_by=('prm', 'year'))

        # On ajoute des données sur le même PRM mais pas sur la même année
        TestDataStoreWithAttribute.set_lc(prm=data_name, value=self.test_data, client_id=self.test_client_id,
                                          attributes_to_set=dict(created_at=created_at_one,
                                                                 last_modified=created_at_one,
                                                                 year=2025),
                                          versionning=True,
                                          versionning_by=('prm', 'year'))

        # On récupére les données data_name en combinant par prm et année (2 entrées)
        data = TestDataStoreWithAttribute.get_lc(data_name, self.test_client_id, combined_versions=True,
                                                 combined_by=('prm', 'year'))
        self.assertEquals(len(data), 2)

        # On récupére les données data_name en combinant par prm et année (1 prm avec 2 entrées)
        data = TestDataStoreWithAttribute.get_many_lc([data_name], self.test_client_id, combined_versions=True,
                                                      combined_by=('prm', 'year'))
        self.assertEquals(len(data), 1)
        self.assertEquals(len(data[data_name]), 2)

        # On récupére les données data_name en combinant par version (toutes à 0)
        data = TestDataStoreWithAttribute.get_lc(data_name, self.test_client_id, combined_versions=True,
                                                 combined_by=('version',))
        self.assertEquals(len(data), 1)

        created_at_two = dt.datetime.now().astimezone(ZoneInfo("Europe/Paris"))
        # Ajout d'une deuxième courbe sur 2024
        TestDataStoreWithAttribute.set_lc(prm=data_name, value=self.test_data_2, client_id=self.test_client_id,
                                          attributes_to_set=dict(created_at=created_at_two,
                                                                 last_modified=created_at_two,
                                                                 year=2024),
                                          versionning=True,
                                          versionning_by=('prm', 'year'))

        # On récupére les données data_name en combinant par prm (2 entrées) mais pour l'année 2024 nous devons avoir les valeur de test_data_2
        data = TestDataStoreWithAttribute.get_lc(data_name, self.test_client_id, combined_versions=True,
                                                 combined_by=('prm', 'year'), order_by=('-created_at',))
        self.assertEquals(len(data), 2)

        data_2024 = next(filter(lambda d: d['year'] == 2024, data), None)
        pd.testing.assert_series_equal(data_2024['data'], self.test_data_2, check_names=False)

    def test_versionning_by(self):
        TestDataStoreWithAttribute.clear_all(self.test_client_id)
        data_name = 'data_name'
        created_at_one = dt.datetime.now().astimezone(ZoneInfo("Europe/Paris"))
        TestDataStoreWithAttribute.set_lc(prm=data_name, value=self.test_data, client_id=self.test_client_id,
                                          attributes_to_set=dict(created_at=created_at_one,
                                                                 last_modified=created_at_one,
                                                                 year=2024),
                                          versionning=True,
                                          versionning_by=('prm',))

        # On ajoute des données sur le même PRM mais pas sur la même année
        TestDataStoreWithAttribute.set_lc(prm=data_name, value=self.test_data_2, client_id=self.test_client_id,
                                          attributes_to_set=dict(created_at=created_at_one,
                                                                 last_modified=created_at_one,
                                                                 year=2025),
                                          versionning=True,
                                          versionning_by=('prm',))

        data = TestDataStoreWithAttribute.get_lc(data_name, self.test_client_id, combined_versions=False, order_by=('-created_at',))
        self.assertEquals(len(data), 2)
        self.assertEquals(sorted([d['version'] for d in data]), [0, 1])
        data = TestDataStoreWithAttribute.get_lc(data_name, self.test_client_id, combined_versions=True, combined_by=('prm',),
                                                 order_by=('-created_at',))
        self.assertEquals(len(data), 1)
        data = TestDataStoreWithAttribute.get_lc(data_name, self.test_client_id, combined_versions=True,
                                                 combined_by=('prm', 'year'),
                                                 order_by=('-created_at',))
        self.assertEquals(len(data), 2)

        # On supprime les données et on versionne par prm et year
        TestDataStoreWithAttribute.clear_all(self.test_client_id)
        data_name = 'data_name'
        created_at_one = dt.datetime.now().astimezone(ZoneInfo("Europe/Paris"))
        TestDataStoreWithAttribute.set_lc(prm=data_name, value=self.test_data, client_id=self.test_client_id,
                                          attributes_to_set=dict(created_at=created_at_one,
                                                                 last_modified=created_at_one,
                                                                 year=2024),
                                          versionning=True,
                                          versionning_by=('prm', 'year'))

        # On ajoute des données sur le même PRM mais pas sur la même année
        TestDataStoreWithAttribute.set_lc(prm=data_name, value=self.test_data_2, client_id=self.test_client_id,
                                          attributes_to_set=dict(created_at=created_at_one,
                                                                 last_modified=created_at_one,
                                                                 year=2024),
                                          versionning=True,
                                          versionning_by=('prm', 'year'))
        TestDataStoreWithAttribute.set_lc(prm=data_name, value=self.test_data, client_id=self.test_client_id,
                                          attributes_to_set=dict(created_at=created_at_one,
                                                                 last_modified=created_at_one,
                                                                 year=2025),
                                          versionning=True,
                                          versionning_by=('prm', 'year'))

        data = TestDataStoreWithAttribute.get_lc(data_name, self.test_client_id, combined_versions=False,
                                                 order_by=('-created_at',))
        self.assertEquals(len(data), 3)
        # 2 versions pour 2024, 1 version pour 2025
        self.assertEquals(sorted([d['version'] for d in data]), [0, 0, 1])
        data = TestDataStoreWithAttribute.get_lc(data_name, self.test_client_id, combined_versions=True,
                                                 combined_by=('prm', 'year'),
                                                 order_by=('-created_at',))
        # 2024 et 2025 combiné
        self.assertEquals(len(data), 2)
        data_2024 = next(filter(lambda d: d['year'] == 2024, data), None)
        pd.testing.assert_series_equal(data_2024['data'], self.test_data_2, check_names=False)
        data_2025 = next(filter(lambda d: d['year'] == 2025, data), None)
        pd.testing.assert_series_equal(data_2025['data'], self.test_data, check_names=False)




import datetime as dt
import logging
from functools import lru_cache
from hashlib import blake2b
from typing import Union, List

import pytz
from django.core.exceptions import ImproperlyConfigured
from django.db import models, transaction
import lz4.frame as lz4
import numpy as np
import pandas as pd
from django.db.models import QuerySet, Max
from django.db.models.base import ModelBase
from django.db.models.signals import class_prepared
from django.dispatch import receiver
from pytz.exceptions import UnknownTimeZoneError
from hostore.utils.timeseries import _localise_date, _localised_now

logger = logging.getLogger(__name__)

# KEYS_ABSTRACT_CLASS = set([field.name for field in TimeseriesChunkStore._meta.get_fields()])
# KEYS_ABSTRACT_CLASS.add('id')
KEYS_ABSTRACT_CLASS = {'id', 'start_ts', 'data', 'dtype', 'updated_at', 'is_deleted', 'chunk_index'}
FROZEN_ATTRS = ('CHUNK_AXIS', 'STORE_TZ', 'STORE_FREQ', 'ALLOW_CLIENT_SERVER_SYNC')

EMPTY_DATA = lz4.compress(np.array([]))
PG_MAX_NAME = 63

class ChunkQuerySet(models.QuerySet):
    """Remplace l'effacement physique par un soft-delete."""
    def delete(self, keep_tracking=False, _disable_sync_safety=False):
        """
        Overridden delete method that handles the tracking of deleted object.

        Raises a ValueError if the user is trying to delete objects with ALLOW_CLIENT_SERVER_SYNC=True and
        keep_tracking=False

        Args:
            keep_tracking:
            _disable_sync_safety: should not be used by user. Disable consistency check between
            ALLOW_CLIENT_SERVER_SYNC and keep_tracking=False
        """
        if not keep_tracking:
            if self.model.ALLOW_CLIENT_SERVER_SYNC and not _disable_sync_safety:
                raise ValueError(f'Trying to delete {self.model.__name__} objects with ALLOW_CLIENT_SERVER_SYNC=True and '
                                 f'keep_tracking=False. This would lead to inconsistencies in client store.')
            super().delete()
        else:
            return super().update(is_deleted=True, data=EMPTY_DATA,
                                  updated_at=_localised_now(timezone_name='UTC'),
                                  )

def _meta_name(model, suffix: str, *, max_len: int = 30) -> str:
    """
    Return a unique, deterministic index name that satisfies common RDBMS
    length limits (≤ *max_len* chars, default 30).

    • Base pattern   "<db_table>_<suffix>"
    • If too long   keep the *prefix* (first part) and append a 4-hex digest
                    so the name stays human-readable **and** unique.

    Example
    -------
    >>> _meta_name(MyModel, "axis")
    'app_mymodel_axis_ab12'
    """
    base = f"{model._meta.db_table}_{suffix}"
    if len(base) <= max_len:
        return base

    # Reserve 5 chars (“_” + 4-char hash) for the uniqueness suffix
    prefix = base[: max_len - 5]
    digest = blake2b(base.encode(), digest_size=2).hexdigest()  # 4 hex chars
    return f"{prefix}_{digest}"

def _tbl_name(app_label: str, model_name: str, sig: str,
              *, max_len: int = PG_MAX_NAME) -> str:
    """
    Renvoie un nom de table <= *max_len*.
    • Patron de base : "<app>_<model>__<signature>"
    • Si trop long  : tronque le préfixe et ajoute un digest sur 8 hex.
    """
    base = f"{app_label}_{model_name.lower()}__{sig}"
    if len(base) <= max_len:
        return base

    # Réserver 9 car. pour "_" + digest(8 hex)
    digest = blake2b(base.encode(), digest_size=4).hexdigest()  # 8 hex
    prefix = base[: max_len - 9]
    return f"{prefix}_{digest}"


class _TCSMeta(ModelBase):
    """
    Métaclasse injectée dans `TimeseriesChunkStore`.

    • Ajoute automatiquement aux sous-classes NON abstraites :
        - UniqueConstraint  (*business_keys, 'chunk_index')
        - Index 'axis'     sur (*business_keys, 'chunk_index')
        - Index 'upd'      sur updated_at
        - Un nom qui fige les propriétés de classe
    """

    def __new__(mcls, name, bases, attrs, **kwargs):
        new_cls = super().__new__(mcls, name, bases, attrs, **kwargs)

        # On ignore la classe abstraite
        if new_cls._meta.abstract:
            return new_cls

        meta = new_cls._meta         # raccourci
        # 1 : UniqueConstraint + indexes
        # 1.1. Keys déclarées par l’utilisateur (hors champs internes)
        business_keys = [
            f.name for f in meta.local_fields          # uniquement ceux ajoutés dans le modèle concret
            if not f.auto_created and f.name not in KEYS_ABSTRACT_CLASS
        ]
        business_keys.sort()
        axis_keys = (*business_keys, "chunk_index")

        # 1.2. UniqueConstraint
        meta.constraints.append(
            models.UniqueConstraint(fields=list(axis_keys), name=_meta_name(new_cls, 'unq'))
        )
        meta.original_attrs["constraints"] = meta.constraints

        # 1.3. INDEXES
        def _has_index(fields) -> bool:
            return any(set(idx.fields) == set(fields) for idx in meta.indexes)

        required = [
            (axis_keys,        "axis"),
            (("updated_at",),  "upd"),
        ]
        for fields, tag in required:
            if not _has_index(fields):
                meta.indexes.append(
                    models.Index(fields=list(fields), name=_meta_name(new_cls, tag))
                )

        # idem pour le détecteur de migrations
        meta.original_attrs["indexes"] = meta.indexes

        # 2 : table signature with class properties
        # 2.1. Construire la signature immuable
        sig = "_".join([
            "ca" + "_".join(new_cls.CHUNK_AXIS),   # ex. cxyear_month
            "f" + new_cls.STORE_FREQ,          # ex. freq1h
            "tz" + new_cls.STORE_TZ.replace('/', '_'),  # ex. tzEurope_Paris
            "sync" if new_cls.ALLOW_CLIENT_SERVER_SYNC else "nosync",
        ])

        # 2.2. Nom de table par défaut : <app>_<model>__<signature>
        default_table = _tbl_name(meta.app_label, name, sig)

        # 2.3. Détecter si l’utilisateur a déjà fixé db_table
        # -----------------------------------------------
        default_django_table = f"{meta.app_label}_{name.lower()}"

        if meta.db_table == default_django_table:
            meta.db_table = default_table
            meta.original_attrs["db_table"] = default_table
        else:
            # Vérif cohérence si db_table a été fourni
            if sig not in meta.db_table:
                raise ImproperlyConfigured(
                    f"`db_table` ({meta.db_table}) ne reflète pas la config "
                    f"immuable ({sig})."
                )
        return new_cls

    def __setattr__(cls, name, value):
        """ block setting any of _FROZEN_ATTRS """
        if name in FROZEN_ATTRS and hasattr(cls, name):
            raise AttributeError(f"{name} est immuable après la 1ʳᵉ migration.")
        super().__setattr__(name, value)


class TimeseriesChunkStore(models.Model, metaclass=_TCSMeta):
    # Partitionnement temporel (null si pas de chunk)
    chunk_index = models.IntegerField()

    # Métadonnées obligatoires
    start_ts = models.DateTimeField()        # premier timestamp inclus
    dtype    = models.CharField(max_length=16)  # ex. 'float64'
    updated_at = models.DateTimeField(auto_now=False)  # synchro
    is_deleted = models.BooleanField(default=False)  # keep trace of deleted objects

    # Données brutes
    data = models.BinaryField()

    # Paramètres haut niveau
    _ALLOWED_CHUNK_AXIS = {('year',), ('year', 'month')}
    CHUNK_AXIS = ('year', 'month')   # Chunking axis for timeseries storage. Configs : ('year',) / ('year', 'month')
    STORE_TZ   = 'Europe/Paris' # Chunking timezone
    STORE_FREQ   = '1h' # Timeseries storage frequency.
    ALLOW_CLIENT_SERVER_SYNC = False # if True, enable the sync features
    CACHED_INDEX_SIZE = 120  # max size for cached date indexes

    _model_keys = None
    _model_td = None
    objects = ChunkQuerySet.as_manager()

    class Meta:
        abstract = True

    @classmethod
    def get_model_keys(cls) -> List[str]:
        """ Returns the list of the keys added to the non abstract class """
        if cls._model_keys is None:
            cls._model_keys = set([field.name for field in cls._meta.get_fields()]) - KEYS_ABSTRACT_CLASS
        return cls._model_keys

    @classmethod
    def _get_model_timedelta(cls) -> pd.Timedelta:
        if cls._model_td is None:
            cls._model_td = pd.to_timedelta(cls.STORE_FREQ)
        return cls._model_td

    # valeurs de chunk autorisées
    def __init_subclass__(cls, **kwargs):
        """Valide les options de classe au moment où la sous-classe est créée."""
        super().__init_subclass__(**kwargs)
        # validate CHUNK_AXIS
        if tuple(getattr(cls, "CHUNK_AXIS", ())) not in cls._ALLOWED_CHUNK_AXIS:
            raise ValueError(
                f"Invalid CHUNK_AXIS {cls.CHUNK_AXIS!r} for model {cls.__name__} — allowed: {cls._ALLOWED_CHUNK_AXIS}"
            )
        # validate STORE_FREQ
        try:
            pd.to_timedelta(cls.STORE_FREQ)
        except ValueError:
            raise ValueError(
                f"Invalid STORE_FREQ {cls.STORE_FREQ!r} for model {cls.__name__}"
            )
        # validate STORE_TZ
        try:
            pytz.timezone(cls.STORE_TZ)
        except UnknownTimeZoneError:
            raise ValueError(
                f"Invalid STORE_TZ {cls.STORE_TZ!r} for model {cls.__name__}"
            )

        # wrap the index rebuild helper with an LRU cache sized per subclass
        cls._cached_index = classmethod(
            lru_cache(maxsize=cls.CACHED_INDEX_SIZE)(cls._cached_index.__func__)
        )

    def delete(self, **kwargs):
        """ Soft-delete : conserve la ligne comme tombstone. """
        self.__class__.objects.filter(id=self.id).delete(**kwargs)

    # ------------------ Sérialisation bas niveau ------------------
    @staticmethod
    def _compress(serie: pd.Series) -> (bytes, np.array):
        arr = serie.to_numpy()
        return lz4.compress(arr.tobytes()), arr

    @classmethod
    def _decompress(cls, row, mem_view=False, return_mode=None) -> pd.Series:
        blob = row.data
        dtype = row.dtype
        if mem_view:
            raw = memoryview(lz4.decompress(blob))
            arr = np.frombuffer(raw, dtype=dtype, like=np.empty(0, dtype=dtype))
        else:
            raw = lz4.decompress(blob)
            arr = np.frombuffer(raw, dtype=dtype)

        if return_mode=='light':
            idx_min, idx_max = cls._bound_index(row, len(arr))
            return idx_min, idx_max, arr
        elif return_mode=='max_idx':
            non_nan_indices = np.where(~np.isnan(arr))[0]
            index_max_non_nan = non_nan_indices[-1] if non_nan_indices.size > 0 else 0

            max_horodate = pd.Timestamp(row.start_ts).tz_convert(cls.STORE_TZ) + index_max_non_nan  * cls._get_model_timedelta()
            return max_horodate
        else:
            idx = cls._rebuild_index(row, len(arr))
            return pd.Series(arr, index=idx)

    # ------------------------------------------------------------------
    # PUBLIC METHODS
    # ------------------------------------------------------------------

    @classmethod
    def last_updated_at(cls, filters: dict = None) -> pd.Timestamp:
        """
        Return the most recent ``updated_at`` timestamp stored in the DB.
        If the table is empty, fall back to 2000-01-01 in the store TZ.

        Parameters
        ----------
        filters   : dict
            Filter the data for the update lookup
        """
        if filters is None:
            filters = dict()

        qs = cls.objects.filter(**filters)
        if qs.exists():
            # return qs.aggregate(last=Max("updated_at"))["last"]
            return qs.order_by('-updated_at').first().updated_at
        else:
            return _localise_date(dt.datetime(2000, 1, 1))

    @classmethod
    def set_ts(cls, attrs: dict, serie: pd.Series,
               update=False, replace=False, bulk_create_batch_size=200):
        """
        Persist a dense time-series in the store.

        Parameters
        ----------
        attrs   : dict
            Business keys identifying the series (version, kind, …).
        serie   : pd.Series
            The data to save.
        update  : bool, default False
            If True, merge the new series with the existing one
            (``combine_first`` logic).
        replace : bool, default False
            If True, delete any existing chunks for the same keys
            before inserting the new series.
        bulk_create_batch_size : bulk batch size for bulk insertion
        Notes
        -----
        * ``update`` and ``replace`` are mutually exclusive.
        * All model business keys must be present in *attrs*.
        * Inserting over existing data with both flags at False raises ``IntegrityError``.
        """
        if update and replace:
            raise ValueError('update and replace are mutually exclusive')

        if (not update) and (not replace) and cls.ALLOW_CLIENT_SERVER_SYNC:
            raise ValueError(f'Trying to use set_ts with model {cls.__name__} '
                             f'without update or replace option while ALLOW_CLIENT_SERVER_SYNC=True.')

        cls._ensure_all_attrs_specified(attrs)
        serie = cls._normalize_serie(serie)
        if serie is None:
            return

        # Enregistrement par chunk
        rows = []
        for sub in cls._chunk(serie):
            if update:
                cls._update_chunk_with_existing(attrs, sub)
            else:
                rows.append(cls._build_row(attrs, sub))
                if len(rows) >= bulk_create_batch_size:
                    cls._bulk_create(rows, bulk_create_batch_size)
                    rows = []

        if replace:
            # we need to hard delete previous related chunks (previous serie may lay on a greater span than new one)
            # only over chunk_index deleted (to allow propagation of deleted index to client store)
            chunk_index_replaced = [r.chunk_index for r in rows]
            qd_attrs = cls.objects.filter(
                **attrs,
            )
            # hard delete part of the serie within existing chunks index : will be replaced in _bulk_create
            # _disable_sync_safety to avoid check raise
            qd_attrs.filter(
                chunk_index__in=chunk_index_replaced
            ).delete(keep_tracking=False, _disable_sync_safety=True)
            # soft delete other chunks : keep trace of deletion
            qd_attrs.delete(keep_tracking=True)

        if not update:
            cls._bulk_create(rows, bulk_create_batch_size)

    @classmethod
    def get_ts(cls, attrs: dict, start: pd.Timestamp=None, end: pd.Timestamp=None, drop_bounds_na=True) -> None | pd.Series:
        """
        Retrieve a time-series matching *attrs*.

        attrs   : dict
            Business keys identifying the series (version, kind, …).
        start   : dt.datetime
            Optional, start of time range to retrieve.
        end   : dt.datetime
            Optional, end of time range to retrieve.
        drop_bounds_na : bool
            Optional, drop nans from result.

        Notes
        -----
        * All model business keys must be present in *attrs*.
        * Requesting non-existing attrs will return None.

        Returns
        -------
        pd.Series | None
            The reconstructed series, or *None* if no chunk matches.
        """
        cls._ensure_all_attrs_specified(attrs)
        qs = cls.objects.filter(**attrs, is_deleted=False).order_by('chunk_index')
        if start or end:
            qs = cls._filter_interval(qs, start, end)

        pieces = []
        for row in qs:
            pieces.append(cls._decompress(row))

        if not pieces:
            return None

        full = pd.concat(pieces)
        serie = cls._finish_serie(full, start, end, drop_bounds_na)

        return serie

    @classmethod
    def set_many_ts(cls, mapping: dict[tuple, pd.Series], keys: tuple[str, ...],
                    bulk_create_batch_size=200):


        """
        Bulk insert many timeseries at once.

        Notes
        -----
        * All model business keys must be present in *attrs*.
        * Trying to insert over existing attrs will raise a django.db.utils.IntegrityError (user must clear matching
        filters upstream)

        Args:
            mapping : {(version_value, kind_value,...): serie}
            keys    : ('version','kind',...)
            bulk_create_batch_size : bulk batch size for bulk insertion

        Returns:
        """
        if cls.ALLOW_CLIENT_SERVER_SYNC:
            raise ValueError(f'Trying to use set_many_ts with model {cls.__name__} '
                             f'while ALLOW_CLIENT_SERVER_SYNC=True.')

        rows = []
        for ktuple, serie in mapping.items():
            attrs = dict(zip(keys, ktuple))
            cls._ensure_all_attrs_specified(attrs)
            serie = cls._normalize_serie(serie)
            if serie is None:
                continue
            for sub in cls._chunk(serie):
                rows.append(cls._build_row(attrs, sub))
                if len(rows) >= bulk_create_batch_size:
                    cls._bulk_create(rows, bulk_create_batch_size)
                    rows = []
        cls._bulk_create(rows, bulk_create_batch_size)

    @classmethod
    def yield_many_ts(cls, filters: dict, start: pd.Timestamp=None, end: pd.Timestamp=None,
                      qs_iterator_chunk_size=200, drop_bounds_na=True):
        """
        Yield (serie, filters_dict) for each available timeseries with filters matching filters.
            - serie will be expressed at STORE_FREQ, STORE_TZ
            - attrs_dict : mapping {model_key: value}

        Args:
            filters: filters of query
            start: start index of timeseries
            end: end index of timeseries
            qs_iterator_chunk_size: size of queryset batch
            drop_bounds_na: drop NaNs
        """
        # On valide seulement les clés fournies
        cls._check_attrs(filters)

        qs = cls.objects.filter(**filters, is_deleted=False).order_by(*(cls.get_model_keys()), 'chunk_index')
        if start or end:
            qs = cls._filter_interval(qs, start, end)

        current_values = None
        buffer_data = []
        min_idx, max_idx = None, None
        def flush():
            if not buffer_data:
                return

            serie = pd.Series(
                index=pd.date_range(start=min_idx, end=max_idx, inclusive='both', freq=cls.STORE_FREQ),
                data=np.concatenate(buffer_data)
            )
            serie = cls._finish_serie(serie, start, end, drop_bounds_na)
            key_dict = dict(zip(cls.get_model_keys(), current_values))
            yield serie, key_dict
            # fixme ? done outside to be allowed to erase _idx
            # buffer_data.clear()
            # min_idx = None
            # max_idx = None

        for row in qs.iterator(chunk_size=qs_iterator_chunk_size):
            values = tuple(getattr(row, k) for k in cls.get_model_keys())
            if current_values is None:
                current_values = values
            elif values != current_values:
                # nouvelle combinaison → on émet la série courante
                yield from flush()
                buffer_data.clear()
                min_idx = None
                max_idx = None

                current_values = values
            _idx_min, _idx_max, data = cls._decompress(row, return_mode='light')
            min_idx = min(min_idx, _idx_min) if min_idx else _idx_min
            max_idx = max(max_idx, _idx_max) if max_idx else _idx_max
            buffer_data.append(data)

        # flush final
        yield from flush()

    @classmethod
    def get_max_horodate(cls, filters: dict, qs_iterator_chunk_size=200):
        """
        Get maximum horodate of timeseries with filters matching filters.

        Args:
            filters: filters of query
        """
        # On valide seulement les clés fournies
        cls._check_attrs(filters)

        qs = cls.objects.filter(**filters, is_deleted=False).order_by(*(cls.get_model_keys()), 'chunk_index')
        max_qs_chunk_index = qs.aggregate(max_chunk=Max('chunk_index'))['max_chunk']
        qs = qs.filter(chunk_index=max_qs_chunk_index)
        current_values = None
        max_idxs = []
        for row in qs.iterator(chunk_size=qs_iterator_chunk_size):
            # max de la série actuelle
            max_idxs.append(cls._decompress(row, return_mode='max_idx'))

        return max(max_idxs) if max_idxs else None

    # ------------------------------------------------------------------
    #  SYNC CLIENT ⇆ SERVER
    # ------------------------------------------------------------------

    @classmethod
    def updates_queryset(cls, since: pd.Timestamp, filters: dict | None = None):
        """Return a queryset of chunks updated after *since*.

        The queryset is ordered by ``updated_at`` and primary key to ensure
        deterministic pagination.
        """
        filters = filters or {}
        return (
            cls.objects
            .filter(**filters, updated_at__gte=since)
            .order_by("updated_at", "pk")
            .values(
                *cls.get_model_keys(),
                "chunk_index",
                "dtype",
                "start_ts",
                "updated_at",
                "is_deleted",
            )
        )

    @classmethod
    def list_updates(
        cls,
        since: pd.Timestamp,
        filters: dict | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[dict]:
        """Convenience helper returning a list from ``updates_queryset``.

        Args:
            since: timestamp lower bound (excluded).
            filters: additional filters on the query.
            limit: maximum number of items to return.
            offset: number of items to skip from the beginning.

        Each dict contains::
            attrs       : full business key dict
            chunk_index : int
            dtype       : str
            start_ts    : dt.datetime
            updated_at  : dt.datetime
            is_deleted  : bool
        """
        qs = cls.updates_queryset(since, filters)
        if offset:
            qs = qs[offset:]
        if limit is not None:
            qs = qs[:limit]

        out = []
        for row in qs:
            out.append({
                "attrs": {k: row[k] for k in cls.get_model_keys()},
                "chunk_index": row["chunk_index"],
                "dtype": row["dtype"],
                "start_ts": row["start_ts"],
                "updated_at": row["updated_at"],
                "is_deleted": row["is_deleted"],
            })
        return out

    @classmethod
    def export_chunks(cls, spec: list[dict]) -> list[tuple]:
        """
        Server-side helper: given *spec* (list of ``{"attrs": ..,
        "chunk_index": ..}``) return raw LZ4 blobs together with their
        metadata (blob_lz4: bytes, attrs: dict, meta: dict).
        No decompression is performed.
        """
        out = []
        for item in spec:
            attrs = item["attrs"]
            idx = item["chunk_index"]
            row = cls.objects.get(**attrs, chunk_index=idx)
            attrs["chunk_index"] = row.chunk_index
            blob = row.data
            meta = {
                "dtype": row.dtype,
                "start_ts": row.start_ts,
                "is_deleted": row.is_deleted,
                "updated_at": row.updated_at,
            }
            out.append((blob, attrs, meta))
        return out

    @classmethod
    def import_chunks(cls, payload: list[tuple]):
        """
        Client-side helper: ingest the list produced by *export_chunks*.
        Each tuple is ``(blob_lz4, attrs_dict, meta_dict)``.
        """
        if not payload:
            return

        # Pre-compute union of meta fields to update using bulk_update later on
        meta_fields = set()
        for _, _, meta in payload:
            meta_fields.update(k for k in meta.keys() if k != "updated_at")

        now = _localised_now(timezone_name="UTC")

        with transaction.atomic():
            attrs_list = [attrs for _, attrs, _ in payload]
            q = models.Q()
            for attrs in attrs_list:
                cls._ensure_all_attrs_specified(attrs, bonus_keys=["chunk_index"])
                q |= models.Q(**attrs)

            existing = {}
            if attrs_list:
                for row in cls.objects.filter(q):
                    key = tuple(sorted((f, getattr(row, f)) for f in attrs_list[0].keys()))
                    existing[key] = row

            rows_to_create = []
            rows_to_update = []
            for blob, attrs, meta in payload:
                key = tuple(sorted(attrs.items()))
                updated_at = meta.get("updated_at")
                if updated_at is not None:
                    updated_at = pd.Timestamp(updated_at)
                    if updated_at.tzinfo is None:
                        updated_at = updated_at.tz_localize("UTC")
                    updated_at = updated_at.to_pydatetime()
                else:
                    updated_at = now
                meta_without_updated = {k: v for k, v in meta.items() if k != "updated_at"}
                if key in existing:
                    row = existing[key]
                    row.data = blob
                    row.updated_at = updated_at
                    for k, v in meta_without_updated.items():
                        setattr(row, k, v)
                    rows_to_update.append(row)
                else:
                    obj = cls(
                        **attrs,
                        data=blob,
                        updated_at=updated_at,
                        **meta_without_updated,
                    )
                    rows_to_create.append(obj)

            if rows_to_create:
                cls.objects.bulk_create(rows_to_create)
            if rows_to_update:
                update_fields = ["data", "updated_at", *meta_fields]
                cls.objects.bulk_update(rows_to_update, update_fields)

    # -- private helpers --

    @classmethod
    def _normalize_serie(cls, serie: pd.Series) -> Union[None, pd.Series]:
        serie = cls._normalize_index(serie)
        if (serie is not None) and (str(serie.dtype) == 'object'):
            serie = serie.apply(float)
        return serie

    @classmethod
    def _normalize_index(cls, serie: pd.Series) -> Union[None, pd.Series]:
        """
        normalize index of serie by reindexing over chunking grid
        Args:
            serie: initial serie

        Returns: transformed serie
        """
        if serie.empty:
            return None

        if not isinstance(serie.index, pd.DatetimeIndex):
            raise ValueError('Index doit être DatetimeIndex.')
        nulls = serie.isnull()
        if nulls.all():
            return None
        # avoid to save null parts in db
        serie = serie[~nulls]
        if serie.index.tz is None:
            logger.warning(f'Saving serie without tz may lead to inconsistent results : localized to STORE_TZ {cls.STORE_TZ}')
            serie = serie.tz_localize(cls.STORE_TZ)
        else:
            serie = serie.tz_convert(cls.STORE_TZ)

        first = serie.index[0]
        last = serie.index[-1]
        if cls.CHUNK_AXIS == ('year',):
            start = first.replace(month=1, day=1, hour=0, minute=0)
            end   = last.replace(day=1, hour=0, minute=0) + pd.offsets.YearEnd() + pd.offsets.Day()
        else:  # ('year','month')
            start = first.replace(day=1, hour=0, minute=0)
            end   = last.replace(day=1, hour=0, minute=0) + pd.offsets.MonthEnd() + pd.offsets.Day()

        new_index = pd.date_range(start=start, end=end, inclusive='left',freq=cls.STORE_FREQ)
        serie = serie.reindex(new_index)

        return serie

    @classmethod
    def _chunk(cls, serie: pd.Series):
        if not cls.CHUNK_AXIS:
            yield serie
        else:
            grouper = serie.groupby([
                getattr(serie.index, ax) for ax in cls.CHUNK_AXIS
            ])
            for _, sub in grouper:
                yield sub

    @classmethod
    def _chunk_index(cls, ts: pd.DatetimeIndex) -> int:
        """ Compute chunk_index value """
        if cls.CHUNK_AXIS == ('year',):
            return ts.year
        # else ('year','month')
        return ts.year * 12 + ts.month - 1

    @classmethod
    def _build_row(cls, attrs: dict, serie: pd.Series):
        """
        Build db object row from serie
        serie must be chunked and normalized
        """
        compressed, arr = cls._compress(serie)
        first_ts = serie.index[0]
        return cls(
            **attrs,
            chunk_index=cls._chunk_index(first_ts),
            start_ts=first_ts,
            dtype=str(arr.dtype),
            data=compressed,
            is_deleted=False,
            updated_at=_localised_now(timezone_name='UTC'),
        )

    @classmethod
    def _update_chunk_with_existing(cls, attrs: dict, serie: pd.Series):
        attributes = {**attrs, 'chunk_index': cls._chunk_index(serie.index[0])}

        with transaction.atomic():
            # combine first with existing
            try:
                row = cls.objects.select_for_update().get(**attributes)
                if row.is_deleted:
                    # avoid using deleted row
                    raise cls.DoesNotExist
                ds_existing = cls._decompress(row)
                ds_new = serie.combine_first(ds_existing)
                row = cls._build_row(attrs, ds_new)
            except cls.DoesNotExist:
                row = cls._build_row(attrs, serie)
            except cls.MultipleObjectsReturned:
                raise ValueError(f'Multiple chunks found for attributes {attributes}')

            # dict pour defaults
            defaults = {
                'start_ts': row.start_ts,
                'dtype': row.dtype,
                'data': row.data,
                'is_deleted': False,
                'updated_at': _localised_now(timezone_name='UTC'),
            }

            cls.objects.update_or_create(
                defaults=defaults,
                **attributes
            )

    @classmethod
    def _bulk_create(cls, rows: list, bulk_create_batch_size: int):
        if not rows:
            return

        with transaction.atomic():
            cls.objects.bulk_create(
                rows,
                batch_size=bulk_create_batch_size,  # optionnel : tuning
            )

    @classmethod
    def _bound_index(cls, row, length: int):
        min_idx = pd.Timestamp(row.start_ts).tz_convert(cls.STORE_TZ)
        max_idx = min_idx + (length - 1) * cls._get_model_timedelta()
        return min_idx, max_idx

    @classmethod
    def _rebuild_index(cls, row, length: int):
        return cls._cached_index(row.start_ts, length)

    @classmethod
    def _cached_index(cls, start_ts, length):
        return pd.date_range(
            start=pd.Timestamp(start_ts).tz_convert(cls.STORE_TZ),
            periods=length,
            freq=cls.STORE_FREQ,
            tz=cls.STORE_TZ,
        )

    @classmethod
    def _filter_interval(cls, qs: QuerySet, start: pd.Timestamp, end: pd.Timestamp):
        if isinstance(start, str):
            start = pd.Timestamp(start, tz=cls.STORE_TZ)
        elif start is None:
            pass
        else:
            start = pd.Timestamp(start).tz_convert(cls.STORE_TZ)

        if isinstance(end, str):
            end = pd.Timestamp(end, tz=cls.STORE_TZ)
        elif end is None:
            pass
        else:
            end = pd.Timestamp(end).tz_convert(cls.STORE_TZ)

        if start:
            qs = qs.filter(chunk_index__gte=cls._chunk_index(start))
        if end:
            qs = qs.filter(chunk_index__lte=cls._chunk_index(end))
        return qs

    @classmethod
    def _ensure_all_attrs_specified(cls, attrs: dict, bonus_keys=None):
        """
        Vérifie que l'utilisateur a renseigné tous les attributs "métiers"
        """
        attrs_keys = set(attrs.keys())
        model_keys = cls.get_model_keys()
        if bonus_keys:
            model_keys = {*model_keys, *bonus_keys}
        if model_keys != attrs_keys:
            raise ValueError(f'Trying to set or get partial attributes {attrs} while full attributes list is {model_keys}')

    @classmethod
    def _finish_serie(cls, serie, start, end, drop_bounds_na):
        if start and end:
            serie = serie.loc[start:end]
        elif start:
            serie = serie.loc[start:]
        elif end:
            serie = serie.loc[:end]

        if drop_bounds_na:
            serie = serie.loc[serie.first_valid_index(): serie.last_valid_index()]

        return serie

    @classmethod
    def _check_attrs(cls, attrs):
        fields_required = []
        for k in attrs.keys():
            if '__' in k:
                k = k.split('__')[0]
            fields_required.append(k)

        bad = set(fields_required) - cls.get_model_keys()
        if bad:
            raise ValueError(f"Unknown attribute(s) {bad}")


# class TestMyTimeseriesChunkStore(TimeseriesChunkStore):
#     version = models.IntegerField()
#     kind_very_long_name_for_testing_purpose = models.CharField(max_length=50)
#     CHUNK_AXIS = ('year', 'month')





# Generated by Django 5.0.2 on 2024-08-21 15:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0006_testdatastorewithattribute'),
    ]

    operations = [
        migrations.CreateModel(
            name='TestTimeseriesStoreWithAttribute',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('data', models.BinaryField(blank=True, null=True)),
                ('version', models.IntegerField(default=0)),
                ('year', models.IntegerField()),
                ('kind', models.CharField(max_length=100)),
            ],
            options={
                'abstract': False,
                'indexes': [models.Index(fields=['version'], name='hostore_tes_version_822f4b_idx')],
                'unique_together': {('version', 'year', 'kind')},
            },
        ),
    ]



from django.test import SimpleTestCase

from hostore.utils.range.range import Range


class RangeTest(SimpleTestCase):

    def test_difference(self):
        # Cas 1
        interval_1 = Range(0, 5)
        interval_2 = Range(1, 4)
        result = interval_1.difference_missing(interval_2, min_delta=0)
        self.assertEqual(result, [])

        # Cas 2
        interval_1 = Range(0, 5)
        interval_2 = Range(-1, 6)
        result = interval_1.difference_missing(interval_2, min_delta=0)
        self.assertEqual(result, [Range(-1, 0), Range(5, 6)])

        # Cas 3
        interval_1 = Range(0, 5)
        interval_2 = Range(-1, 3)
        result = interval_1.difference_missing(interval_2, min_delta=0)
        self.assertEqual(result, [Range(-1, 0)])

        # Cas 4
        interval_1 = Range(0, 5)
        interval_2 = Range(3, 6)
        result = interval_1.difference_missing(interval_2, min_delta=0)
        self.assertEqual(result, [Range(5, 6)])

        # Cas 5
        interval_1 = Range(0, 5)
        interval_2 = Range(8, 9)
        result = interval_1.difference_missing(interval_2, min_delta=0)
        self.assertEqual(result, [Range(5, 9)])

        # Cas 6
        interval_1 = Range(0, 5)
        interval_2 = Range(-2, -1)
        result = interval_1.difference_missing(interval_2, min_delta=0)
        self.assertEqual(result, [Range(-2, 0)])

        # Cas 7
        interval_1 = Range(0, 5)
        interval_2 = Range(-0.5, 6)
        result = interval_1.difference_missing(interval_2, min_delta=1)
        self.assertEqual(result, [Range(5, 6)])

        # Cas 8
        interval_1 = Range(0, 5)
        interval_2 = Range(0, 6)
        result = interval_1.difference_missing(interval_2, min_delta=1)
        self.assertEqual(result, [Range(5, 6)])

    def test_intersection(self):
        # Cas 1
        interval_1 = Range(0, 5)
        interval_2 = Range(1, 4)
        result = interval_1.intersection(interval_2, min_delta=0)
        self.assertEqual(result, Range(1, 4))

        # Cas 2
        interval_1 = Range(0, 5)
        interval_2 = Range(4, 5)
        result = interval_1.intersection(interval_2, min_delta=2)
        self.assertEqual(result, None)

        # Cas 3
        interval_1 = Range(-1, 3)
        interval_2 = Range(4, 6)
        result = interval_1.intersection(interval_2, min_delta=0)
        self.assertEqual(result, None)

    def test_combine(self):
        # Cas 1
        combined = Range.combine([Range(0, 5), Range(-3, -2), Range(6, 7)])
        self.assertEqual(combined, [Range(-3, -2), Range(0, 5) , Range(6, 7)])

        # Cas 2
        combined = Range.combine([Range(-3, 5), Range(-4, -2), Range(6, 7)])
        self.assertEqual(combined, [Range(-4, 5), Range(6, 7)])

        # Cas 3
        combined = Range.combine([Range(-3, 6), Range(-4, -2), Range(6, 7)])
        self.assertEqual(combined, [Range(-4, 7)])

        # Cas 3
        combined = Range.combine([])
        self.assertEqual(combined, [])

    def test_hash(self):
        r = Range(0, 5)
        self.assertTrue(isinstance(r.__hash__(), int))



# Generated by Django 5.0.2 on 2024-05-29 08:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0005_alter_testdatastore_prm'),
    ]

    operations = [
        migrations.CreateModel(
            name='TestDataStoreWithAttribute',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('client_id', models.IntegerField()),
                ('prm', models.CharField(max_length=70)),
                ('last_modified', models.DateTimeField(auto_now=True)),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True)),
                ('data', models.BinaryField(blank=True, null=True)),
                ('version', models.IntegerField(default=0)),
                ('year', models.IntegerField()),
            ],
            options={
                'abstract': False,
                'indexes': [models.Index(fields=['prm', 'client_id'], name='hostore_tes_prm_93d669_idx')],
                'unique_together': {('prm', 'client_id', 'year', 'created_at')},
            },
        ),
    ]



# Generated by Django 5.2.4 on 2025-07-18 14:22

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0008_remove_testtimeseriesstorewithattribute_hostore_tes_version_822f4b_idx_and_more'),
    ]

    operations = [
        migrations.DeleteModel(
            name='TestDataStore',
        ),
        migrations.DeleteModel(
            name='TestDataStoreWithAttribute',
        ),
        migrations.DeleteModel(
            name='TestTimeseriesStoreWithAttribute',
        ),
    ]



# Generated by Django 5.0.2 on 2024-08-21 16:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0007_testtimeseriesstorewithattribute'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='testtimeseriesstorewithattribute',
            name='hostore_tes_version_822f4b_idx',
        ),
        migrations.AlterUniqueTogether(
            name='testtimeseriesstorewithattribute',
            unique_together={('year', 'kind')},
        ),
        migrations.AddIndex(
            model_name='testtimeseriesstorewithattribute',
            index=models.Index(fields=['data'], name='hostore_tes_data_9fd074_idx'),
        ),
        migrations.RemoveField(
            model_name='testtimeseriesstorewithattribute',
            name='version',
        ),
    ]



# tests/test_timeseries_chunk_store_load.py
import time
import itertools as it
from unittest import skip

import numpy as np
import pandas as pd
from django.db import models, connection
from django.test import TransactionTestCase

from hostore.models import TimeseriesChunkStore


# ---------------------------------------------------------------------------
# 1)  Deux magasins de test
# ---------------------------------------------------------------------------

class LoadStoreMonth15(TimeseriesChunkStore):
    """Chunk (year, month)"""
    version = models.IntegerField()
    kind    = models.CharField(max_length=20)
    CHUNK_AXIS = ("year", "month")
    STORE_FREQ   = '15min'

    class Meta:
        app_label = "ts_inline"
        managed   = True
        constraints = [models.UniqueConstraint(fields=["version", "kind", "chunk_index"], name='hostore_LoadStoreMonth15_unq'), ]
        indexes = [
            models.Index(fields=['version', 'kind', 'chunk_index']),
            models.Index(fields=['updated_at']),
        ]

class LoadStoreYear(TimeseriesChunkStore):
    """Chunk (year)"""
    version = models.IntegerField()
    kind    = models.CharField(max_length=20)
    CHUNK_AXIS = ("year",)

    class Meta:
        app_label = "ts_inline"
        managed   = True
        constraints = [models.UniqueConstraint(fields=["version", "kind", "chunk_index"], name='hostore_LoadStoreYear_unq'), ]
        indexes = [
            models.Index(fields=['version', 'kind', 'chunk_index']),
            models.Index(fields=['updated_at']),
        ]

# ---------------------------------------------------------------------------
# 2)  Base abstraite « load »
# ---------------------------------------------------------------------------
class BaseLoadTsStoreTestCase(TransactionTestCase):
    """
    Classe mère pour les tests de charge.
    Surcharger : STORE_CLASS, N_SERIES, SERIES_LEN_H.
    """
    __unittest_skip__ = True

    STORE_CLASS   = None            # → LoadStoreMonth / LoadStoreYear
    N_SERIES      = 50              # combien de courbes
    SERIES_LEN_H  = 24 * 365 * 2    # taille de chaque courbe (heures)
    RANDOM_SEED   = 0

    # ---------- infra ----------

    def _ensure_tables(self):
        existing = connection.introspection.table_names()
        with connection.schema_editor(atomic=False) as se:
            if self.STORE_CLASS._meta.db_table not in existing:
                se.create_model(self.STORE_CLASS)
        self.STORE_CLASS.objects.all().delete()

    def setUp(self):
        self._ensure_tables()

    # ---------- helpers ----------

    @classmethod
    def _make_serie(cls, start_ts, seed_offset):
        np.random.seed(cls.RANDOM_SEED + seed_offset)
        idx = pd.date_range(
            start=start_ts, periods=cls.SERIES_LEN_H,
            freq=cls.STORE_CLASS.STORE_FREQ, tz=cls.STORE_CLASS.STORE_TZ
        )
        return pd.Series(np.random.randn(cls.SERIES_LEN_H), index=idx)

    # ---------- test de charge ----------

    def test_bulk_insert_and_read(self):
        """
        1. Génère N_SERIES séries aléatoires.
        2. Les insère en bulk.
        3. Relit chaque courbe et vérifie l’intégrité.
        4. Mesure le temps global (simple indicateur).
        """
        Store = self.STORE_CLASS

        # 1) Construction du mapping à insérer
        mapping = {}
        for i in range(self.N_SERIES):
            attrs = (i, f"k{i:03d}")   # (version, kind)
            start = f"2020-01-01"
            mapping[attrs] = self._make_serie(start, i)

        # 2) Insertion bulk
        t0 = time.perf_counter()
        Store.set_many_ts(mapping, keys=("version", "kind"))
        insert_dur = time.perf_counter() - t0

        # 3) Lecture / contrôle get
        t1 = time.perf_counter()
        for (v, k), src in mapping.items():
            dst = Store.get_ts({"version": v, "kind": k})
            # assert_series_equal(src, dst)
        read_dur = time.perf_counter() - t1

        # 4) Lecture / contrôle yield_many
        t1 = time.perf_counter()
        data = Store.yield_many_ts({})
        data = list(data)
        read_many_dur = time.perf_counter() - t1

        # 4) Affiche un petit récap (pas d’assert stricte : ajustable)
        print(
            # f"\n[LOAD] {self.N_SERIES} séries × {self.SERIES_LEN_H/24}j "
            f"→ insert {insert_dur:.2f}s, read {read_dur:.2f}s, read many {read_many_dur:.2f}s "
            f"({Store.__name__})"
        )

@skip
class LoadMonth_heavy(BaseLoadTsStoreTestCase):
    """
    ./manage.py test hostore.tests.test_chunk_timeserie_store_perf.LoadMonth_heavy
    """
    __unittest_skip__ = False
    STORE_CLASS = LoadStoreMonth15
    N_SERIES    = 1000
    SERIES_LEN_H = 24 * 365 * 3 * 4

@skip
class LoadYear_light(BaseLoadTsStoreTestCase):
    """
    ./manage.py test hostore.tests.test_chunk_timeserie_store_perf.LoadYear_light
    """
    __unittest_skip__ = False
    STORE_CLASS = LoadStoreYear
    N_SERIES    = 20
    SERIES_LEN_H = 24 * 365 * 5


import datetime as dt
import numpy as np
import pandas as pd
from django.db import models, IntegrityError
from django.test import TransactionTestCase

from hostore.models import TimeseriesChunkStore
from hostore.utils.timeseries import ts_combine_first, _localise_date_interval
from hostore.utils.utils_test import TempTestTableHelper


class TestStoreChunkYearMonth(TimeseriesChunkStore):
    version = models.IntegerField()
    kind_very_long_name_for_testing_purpose = models.CharField(max_length=50)
    CHUNK_AXIS = ('year', 'month')

    class Meta:
        app_label = "ts_inline"
        managed = True

class TestStoreChunkYearMonthUTC(TimeseriesChunkStore):
    version = models.IntegerField()
    kind_very_long_name_for_testing_purpose = models.CharField(max_length=50)
    CHUNK_AXIS = ('year', 'month')
    STORE_TZ = 'UTC'

    class Meta:
        app_label = "ts_inline"
        managed = True

class TestStoreChunkYear(TimeseriesChunkStore):
    version = models.IntegerField()
    kind_very_long_name_for_testing_purpose = models.CharField(max_length=50)
    CHUNK_AXIS = ('year', )

    class Meta:
        app_label = "ts_inline"
        managed = True


class TestStoreChunkYearMonthNoAttrs(TimeseriesChunkStore):
    CHUNK_AXIS = ('year', 'month')

    class Meta:
        app_label = "ts_inline"
        managed = True


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def assert_series_equal(s1, s2, **kwargs):
    tz_check = 'UTC'
    s1 = s1.dropna()
    s2 = s2.dropna()
    s1 = s1.tz_convert(tz_check)
    s2 = s2.tz_convert(tz_check)
    try:
        pd.testing.assert_series_equal(s1, s2, **kwargs, check_freq=False)
    except AssertionError as e:
        diff = s1-s2
        diff = diff.loc[diff != 0]
        print('assertion error with series diff')
        print(diff)
        raise e

# ---------------------------------------------------------------------------
# Tests principaux
# ---------------------------------------------------------------------------

class BaseTimeseriesChunkStoreTestCase(TransactionTestCase, TempTestTableHelper):
    __unittest_skip__ = True

    test_table = None
    year_count_expected = None
    series_na = None
    drop_series_na = False
    no_user_fields = False  # special case with table without any attributes
    input_tz = 'Europe/Paris'

    # -------------------------------------------------------------------
    # création des tables à chaque test
    # -------------------------------------------------------------------

    def setUp(self):
        # Flush de TransactionTestCase supprime nos tables dynamiques.
        # On les recrée si nécessaire avant chaque test.
        self._ensure_tables(delete_kw=dict(keep_tracking=False))

    @classmethod
    def make_series(cls, start, periods, freq="1h", seed=0, full_nan=False):
        """Génère une série aléatoire de longueur `periods`."""
        rng = pd.date_range(start=start, periods=periods, freq=freq, tz=cls.input_tz)
        if full_nan:
            return pd.Series(None, index=rng)
        np.random.seed(seed)
        ds = pd.Series(np.random.randn(periods), index=rng)
        if cls.series_na is not None:
            for imin, imax in cls.series_na:
                ds.iloc[imin:imax] = None

        if cls.drop_series_na:
            ds.dropna(inplace=True)
        return ds

    @classmethod
    def make_attrs(cls, orig_attrs):
        if cls.no_user_fields:
            return {}
        else:
            return orig_attrs

    # -------------------------------------------------------------------
    # Scénarios
    # -------------------------------------------------------------------

    def test_set_and_get(self):
        # serie_a only
        serie_a = self.make_series("2020-01-01", 24 * 365)
        attrs = self.make_attrs({"version": 1, "kind_very_long_name_for_testing_purpose": "A"})
        self.test_table.set_ts(attrs, serie_a)
        got = self.test_table.get_ts(attrs)

        # test index metadatas
        self.assertEqual(str(got.index.tz), self.test_table.STORE_TZ)
        self.assertIsInstance(got, pd.Series)
        self.assertIsInstance(got.index, pd.DatetimeIndex)
        self.assertEqual(got.index.freq, self.test_table.STORE_FREQ)

        # test content
        assert_series_equal(got, serie_a)
        self.assertGreaterEqual(self.test_table.objects.filter(**attrs).count(), self.year_count_expected)

        # serie_a + serie_b
        if not self.no_user_fields:
            serie_b = self.make_series("2020-01-01", 24 * 345)
            attrs = self.make_attrs({"version": 1, "kind_very_long_name_for_testing_purpose": "B"})
            self.test_table.set_ts(attrs, serie_b)
            got = self.test_table.get_ts(attrs)
            assert_series_equal(got, serie_b)

            self.assertEqual(self.test_table.get_max_horodate({"version": 1}).isoformat(),
                             serie_a[serie_a.notna()].index.max().tz_convert(self.test_table.STORE_TZ).isoformat())

        # try re set other
        serie_c = self.make_series("2020-01-01", 24 * 365)
        with self.assertRaises(IntegrityError):
            self.test_table.set_ts(attrs, serie_c)

        # nonexistent
        if not self.no_user_fields:
            got = self.test_table.get_ts({"version": 1, "kind_very_long_name_for_testing_purpose": "nonexistent"})
            self.assertEqual(got, None)

    def test_set_and_get_underconstrained(self):
        # "under constrained" request - secured by _ensure_all_attrs_specified
        if self.no_user_fields:
            return

        serie_a = self.make_series("2020-01-01", 24 * 365)
        attrs = self.make_attrs({"version": 1})
        with self.assertRaises(ValueError):
            self.test_table.set_ts(attrs, serie_a)

        with self.assertRaises(ValueError):
            got = self.test_table.get_ts(attrs)

    def test_set_and_get_full_nan(self):
        serie = self.make_series("2020-01-01", 24 * 365, full_nan=True)
        attrs = self.make_attrs({"version": 1, "kind_very_long_name_for_testing_purpose": "Anan"})
        self.test_table.set_ts(attrs, serie)
        got = self.test_table.get_ts(attrs)
        self.assertEqual(got, None)
        self.assertEqual(self.test_table.get_max_horodate(attrs), None)

    def test_range_filter(self):
        serie = self.make_series("2019-01-01", 24 * 365)
        attrs = self.make_attrs({"version": 3, "kind_very_long_name_for_testing_purpose": "C"})
        self.test_table.set_ts(attrs, serie)
        start, end = _localise_date_interval(dt.date(2019, 6, 1), dt.date(2019, 6, 2))
        sub = self.test_table.get_ts(attrs, start=start, end=end)
        expected = serie[start:end]
        assert_series_equal(sub, expected)

    def test_update_and_replace(self):
        """
        ./manage.py test hostore.tests.test_chunk_timeserie_store.BaseTimeseriesChunkStoreTestCase.test_update_and_replace
        """
        s1 = self.make_series("2022-01-01", 365*24)
        s2 = self.make_series("2022-06-01", 380*24, seed=42)
        s3 = self.make_series("2022-06-01", 380*24, seed=42)
        self._abstract_test_update_and_replace(s1, s2, s3)

    def test_update_and_replace_discontinuous(self):
        s1 = self.make_series("2022-01-01", 24 * 10)
        s2 = self.make_series("2022-01-15", 24 * 5, seed=42)
        s3 = self.make_series("2022-06-01", 24 * 280, seed=42)
        self._abstract_test_update_and_replace(s1, s2, s3)

    def _abstract_test_update_and_replace(self, s1, s2, s3):
        """
        s1 : first set
        s2 : update s1
        s3 : replace s2
        """
        # other serie
        attrs = self.make_attrs({"version": 4, "kind_very_long_name_for_testing_purpose": "D"})
        attrs_ot = {"version": 4, "kind_very_long_name_for_testing_purpose": "other"}
        s4 = self.make_series("2022-06-01", 380*24, seed=42)

        # other
        if not self.no_user_fields:
            self.test_table.set_ts(attrs_ot, s4)
            assert_series_equal(self.test_table.get_ts(attrs_ot), s4)

        self.test_table.set_ts(attrs, s1)
        assert_series_equal(self.test_table.get_ts(attrs), s1)
        horomax = s1[s1.notna()].index.max().tz_convert(self.test_table.STORE_TZ)
        self.assertEqual(self.test_table.get_max_horodate(attrs).isoformat(),
                         horomax.isoformat())

        self.test_table.set_ts(attrs, s2, update=True)
        assert_series_equal(self.test_table.get_ts(attrs), ts_combine_first([s2, s1]))
        self.assertEqual(self.test_table.get_max_horodate(attrs).isoformat(),
                         max(
                             horomax,
                             s2[s2.notna()].index.max().tz_convert(self.test_table.STORE_TZ)
                         ).isoformat())
        self.test_table.set_ts(attrs, s3, replace=True)
        assert_series_equal(self.test_table.get_ts(attrs), s3)

        # other
        if not self.no_user_fields:
            assert_series_equal(self.test_table.get_ts(attrs_ot), s4)

    def test_update_and_replace_simple(self):
        """
        ./manage.py test hostore.tests.test_chunk_timeserie_store.BaseTimeseriesChunkStoreTestCase.test_update_and_replace_simple
        """
        s1 = self.make_series("2022-01-01", 365*24)
        s2 = self.make_series("2022-06-01", 380*24, seed=42)
        s3 = self.make_series("2022-06-01", 381*24, seed=43)
        s4 = self.make_series("2022-06-01", 420*24, seed=44)
        s5 = self.make_series("2026-06-01", 365*24, seed=44)
        attrs = self.make_attrs({"version": 4, "kind_very_long_name_for_testing_purpose": "D2"})

        self.test_table.set_ts(attrs, s1)
        self.test_table.set_ts(attrs, s2, update=True)
        assert_series_equal(self.test_table.get_ts(attrs), ts_combine_first([s2, s1]))
        self.test_table.set_ts(attrs, s3, update=True)
        assert_series_equal(self.test_table.get_ts(attrs), ts_combine_first([s3, s2, s1]))
        self.test_table.set_ts(attrs, s4, update=True)
        assert_series_equal(self.test_table.get_ts(attrs), ts_combine_first([s4, s3, s2, s1]))
        self.test_table.set_ts(attrs, s5, update=True)
        assert_series_equal(self.test_table.get_ts(attrs), ts_combine_first([s5, s4, s3, s2, s1]))

    def test_set_many_ts(self):
        if self.no_user_fields:
            return
        mapping = {
            (5, "E"): self.make_series("2023-01-01", 24),
            (5, "F"): self.make_series("2023-02-01", 24 * 2),
        }
        self.test_table.set_many_ts(mapping, keys=("version", "kind_very_long_name_for_testing_purpose"))
        for (v, k), serie in mapping.items():
            got = self.test_table.get_ts({"version": v, "kind_very_long_name_for_testing_purpose": k})
            assert_series_equal(got, serie)

        with self.assertRaises(IntegrityError):
            self.test_table.set_many_ts(mapping, keys=("version", "kind_very_long_name_for_testing_purpose"))
    def test_yield_ts(self):
        if self.no_user_fields:
            return

        mapping = {
            (6, "G"): self.make_series("2024-01-01", 24*390),
            (6, "H"): self.make_series("2024-01-01", 24*390),
            (6, "I"): self.make_series("2023-01-01", 24*366),
            (5, "G"): self.make_series("2028-01-01", 24*15),
        }
        self.test_table.set_many_ts(mapping, keys=("version", "kind_very_long_name_for_testing_purpose"))
        seen = {
            (key_dict['version'], key_dict['kind_very_long_name_for_testing_purpose']): serie
            for serie, key_dict in self.test_table.yield_many_ts({"version": 6})
        }
        self.assertEqual(set(seen.keys()), {(6, "G"), (6, "H"), (6, "I")})
        for key, serie in seen.items():
            assert_series_equal(serie, mapping[key])

    def test_invalid_calls(self):
        serie = self.make_series("2025-01-01", 24)
        with self.assertRaises(ValueError):
            self.test_table.set_ts({"version": 7, "kind_very_long_name_for_testing_purpose": "I"}, serie, update=True, replace=True)

        bad = pd.Series([1, 2, 3])
        with self.assertRaises(ValueError):
            self.test_table.set_ts({"version": 8, "kind_very_long_name_for_testing_purpose": "J"}, bad)


class TestTimeseries_1ChunkTestCase(BaseTimeseriesChunkStoreTestCase):
    """
    Test chunk Year
    """
    __unittest_skip__ = False
    test_table = TestStoreChunkYear
    year_count_expected = 1

class TestTimeseries_2ChunkTestCase(BaseTimeseriesChunkStoreTestCase):
    """
    Test chunk Year Month
    """
    __unittest_skip__ = False
    test_table = TestStoreChunkYearMonth
    year_count_expected = 12

class TestTimeseries_2Chunk_NoAttrs_TestCase(BaseTimeseriesChunkStoreTestCase):
    """
    Test chunk Year Month
    """
    __unittest_skip__ = False
    test_table = TestStoreChunkYearMonthNoAttrs
    year_count_expected = 12
    no_user_fields = True

class TestTimeseries_2ChunkUTCTestCase(BaseTimeseriesChunkStoreTestCase):
    """
    Test chunk Year Month
    Table en UTC
    """
    __unittest_skip__ = False
    test_table = TestStoreChunkYearMonthUTC
    year_count_expected = 12

class TestTimeseries_1Chunk_NanTestCase(BaseTimeseriesChunkStoreTestCase):
    """
    Test chunk Year
    Series input avec nan
    """
    __unittest_skip__ = False
    test_table = TestStoreChunkYear
    year_count_expected = 1
    series_na = [
        [24*20, 24*60],
        [24*150, 24*200],
        [24*205, 24*220],
    ]

class TestTimeseries_2Chunk_NanTestCase(BaseTimeseriesChunkStoreTestCase):
    """
    Test chunk Year Month
    Series input avec nan
    """
    __unittest_skip__ = False
    test_table = TestStoreChunkYearMonth
    year_count_expected = 12
    series_na = [
        [24*20, 24*60],
        [24*150, 24*200],
        [24*205, 24*220],
    ]

class TestTimeseries_2Chunk_NanUtcTestCase(BaseTimeseriesChunkStoreTestCase):
    """
    Test chunk Year Month
    Series input avec nan, en TZ UTC
    """
    __unittest_skip__ = False
    test_table = TestStoreChunkYearMonth
    year_count_expected = 12
    series_na = [
        [24 * 20, 24 * 60],
        [24 * 150, 24 * 200],
        [24 * 201, 24 * 220],
    ]
    input_tz='UTC'




# Generated by Django 5.0.2 on 2024-05-23 15:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0004_testdatastore_created_at'),
    ]

    operations = [
        migrations.AlterField(
            model_name='testdatastore',
            name='prm',
            field=models.CharField(max_length=70),
        ),
    ]
